from pathlib import Path
from typing import Any, List, Tuple, Union

from prefect.core import Task
from prefect.utilities.tasks import defaults_from_attrs
from python_http_client.client import Response


class SendEmail(Task):
    """
    A task for sending an email via Twilio SendGrid.

    Args:
        - from_email (str): The email address of the sender; defaults to notifications@prefect.io
        - to_emails (Union[str, Tuple[str, str], List[str], List[Tuple[str, str]]]):
            The email address of the recipient(s); can also be provided at runtime.
            Refer to [SendGrid-Python](https://github.com/sendgrid/sendgrid-python) for specifics.
        - subject (str, optional): The subject of the email; can also be provided at runtime
        - html_content (str): The html body of the email; can also be provided at runtime
        - category (Union[str, List[str]], optional): The category/categories to use for the email;
            can also be provided at runtime
        - attachment_file_path (Union[str, Path], optional): The file path of the email attachment;
            can also be provided at runtime
        - sendgrid_api_key (str): The SendGrid API key used for authentication;
            can also be provided at runtime, which is preferred since a secret can be used
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(
        self,
        from_email: str = "notifications@prefect.io",
        to_emails: Union[str, Tuple[str, str], List[str], List[Tuple[str, str]]] = None,
        subject: str = None,
        html_content: str = None,
        category: Union[str, List[str]] = None,
        attachment_file_path: Union[str, Path] = None,
        sendgrid_api_key: str = None,
        **kwargs: Any
    ):
        self.from_email = from_email
        self.to_emails = to_emails
        self.subject = subject
        self.html_content = html_content
        self.category = category
        self.attachment_file_path = attachment_file_path
        self.sendgrid_api_key = sendgrid_api_key
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "from_email",
        "to_emails",
        "subject",
        "html_content",
        "category",
        "attachment_file_path",
        "sendgrid_api_key",
    )
    def run(
        self,
        from_email: str = "notifications@prefect.io",
        to_emails: Union[str, Tuple[str, str], List[str], List[Tuple[str, str]]] = None,
        subject: str = None,
        html_content: str = None,
        category: Union[str, List[str]] = None,
        attachment_file_path: Union[str, Path] = None,
        sendgrid_api_key: str = None,
    ) -> Response:
        """
        Run message which sends an email via SendGrid.

        Args:
            - from_email (str): The email address of the sender;
                defaults to the one provided at initialization
            - to_emails (Union[str, Tuple[str, str], List[str], List[Tuple[str, str]]]):
                The email address of the recipient(s); defaults to the one provided at initialization.
                Refer to [SendGrid-Python](https://github.com/sendgrid/sendgrid-python) for specifics.
            - subject (str, optional): The subject of the email;
                defaults to the one provided at initialization
            - html_content (str): The html body of the email;
                defaults to the one provided at initialization
            - category (Union[str, List[str]], optional): The category/categories to use for the email;
                defaults to those provided at initialization
            - attachment_file_path (Union[str, Path], optional): The file path of the email attachment;
                defaults to the one provided at initialization
            - sendgrid_api_key (str, optional): The SendGrid API key used for authentication;
                if not provided here, will use the value provided at initialization

        Raises:
            - ValueError: if no API key is provided

        Returns:
            - python_http_client.client.Response:
                A [Python-HTTP-Client](https://github.com/sendgrid/python-http-client) object
                indicating the status of the response
        """

        # Based on the SendGrid example use-case code here:
        # https://github.com/sendgrid/sendgrid-python/blob/aa39f715a061f0de993811faea0adb8223657d01/use_cases/attachment.md

        if not sendgrid_api_key:
            raise ValueError("A SendGrid API key must be provided.")

        import base64
        import mimetypes
        from sendgrid.helpers.mail import (
            Attachment,
            Category,
            Disposition,
            FileContent,
            FileName,
            FileType,
            Mail,
        )
        from sendgrid import SendGridAPIClient

        message = Mail(
            from_email=from_email,
            to_emails=to_emails,
            subject=subject,
            html_content=html_content or "\n",
        )

        if category:
            if not isinstance(category, list):
                category = [category]
            message.category = [Category(str(c)) for c in category]

        if attachment_file_path:
            with open(attachment_file_path, "rb") as f:
                data = f.read()
                f.close()
            encoded = base64.b64encode(data).decode()

            guessed_type, content_encoding = mimetypes.guess_type(
                attachment_file_path, strict=True
            )

            attachment = Attachment()
            attachment.file_content = FileContent(encoded)
            attachment.file_type = FileType(guessed_type)
            attachment.file_name = FileName(Path(attachment_file_path).name)
            attachment.disposition = Disposition("attachment")
            message.attachment = attachment

        sendgrid_client = SendGridAPIClient(sendgrid_api_key)
        response = sendgrid_client.send(message)

        return response



from prefect import Task
import requests
from prefect.utilities.tasks import defaults_from_attrs
from typing import Any


class CreateCard(Task):
    """
    Task for creating a card on Trello, given the list to add it to.

    Args:
      - list_id (str): the id of the list to add the new item.
      - card_name (str, optional): the title of the card to add
      - card_info (str, optional): the description for the back of the card
      - **kwargs (dict, optional): additional arguments to pass to the Task constructor
    """

    def __init__(
        self,
        list_id: str = None,
        card_name: str = None,
        card_info: str = None,
        **kwargs: Any
    ):
        self.list_id = list_id
        self.card_name = card_name
        self.card_info = card_info

        super().__init__(**kwargs)

    @defaults_from_attrs("list_id", "card_name", "card_info")
    def run(
        self,
        list_id: str = None,
        card_name: str = None,
        card_info: str = None,
        trello_api_key: str = "TRELLO_API_KEY",
        trello_server_token: str = "TRELLO_SERVER_TOKEN",
    ) -> None:
        """
        Task run method.

        Args:
          - list_id (str): the id of the list to add the new item.
          - card_name (str, optional): the title of the card to add
          - card_info (str, optional): the description for the back of the card
          - trello_api_key (str): the name of the Prefect Secret where you've stored your API key
          - trello_server_token (str): the name of the Prefect Secret
          where you've stored your server token

        Returns:
          - Status code of POST request (200 for success)
        """

        if list_id is None:
            raise ValueError(
                """A list ID must be provided. See
https://developer.atlassian.com/cloud/trello/rest/api-group-boards/#api-boards-id-lists-get
for details."""
            )

        if trello_api_key is None:
            raise ValueError(
                "An API key must be provided. See https://trello.com/app-key while logged in."
            )

        if trello_server_token is None:
            raise ValueError(
                """A server token must be provided.
Click the 'Token' link on https://trello.com/app-key to generate."""
            )

        url = "https://api.trello.com/1/cards/"
        params = {
            "key": trello_api_key,
            "token": trello_server_token,
            "idList": list_id,
            "name": card_name,
            "desc": card_info,
        }
        response = requests.post(url, data=params)

        return response



"""
This is a collection of tasks to interact with a Cube.js or Cube Cloud environment.
"""

try:
    from prefect.tasks.cubejs.cubejs_tasks import CubeJSQueryTask
except ImportError as err:
    raise ImportError(
        'prefect.tasks.cubejs` requires Prefect to be installed with the "cubejs" extra.'
    ) from err



from time import sleep
from typing import List, Tuple

import requests
import prefect


# dbt Cloud Trigger Job API -> https://docs.getdbt.com/dbt-cloud/api-v2#operation/triggerRun
__DBT_CLOUD_TRIGGER_JOB_API_ENDPOINT_V2 = (
    "https://{apiDomain}/api/v2/accounts/{accountId}/jobs/{jobId}/run/"
)

# dbt Cloud Get Run API -> https://docs.getdbt.com/dbt-cloud/api-v2#operation/getRunById
__DBT_CLOUD_GET_RUN_API_ENDPOINT_V2 = (
    "https://{apiDomain}/api/v2/accounts/{accountId}/runs/{runId}/"
)

# dbt Cloud List Run Artifacts API ->
#   https://docs.getdbt.com/dbt-cloud/api-v2#operation/listArtifactsByRunId
__DBT_CLOUD_LIST_RUN_ARTIFACTS_ENDPOINT_V2 = (
    "https://{apiDomain}/api/v2/accounts/{accountId}/runs/{runId}/artifacts/"
)

# dbt Cloud Get Run Artifact API ->
#   https://docs.getdbt.com/dbt-cloud/api-v2#operation/getArtifactsByRunId
__DBT_CLOUD_GET_RUN_ARTIFACT_ENDPOINT_V2 = (
    "https://{apiDomain}/api/v2/accounts/{accountId}/runs/{runId}/artifacts/{path}"
)

USER_AGENT_HEADER = {"user-agent": f"prefect-{prefect.__version__}"}


class DbtCloudBaseException(Exception):
    """Base exception for all dbt Cloud errors"""

    pass


class TriggerDbtCloudRunFailed(DbtCloudBaseException):
    """Raised when triggering a dbt job run fails"""

    pass


class GetDbtCloudRunFailed(DbtCloudBaseException):
    """Raised when details for a dbt Cloud job run cannot be retrieved"""

    pass


class DbtCloudRunFailed(DbtCloudBaseException):
    """Raised when a dbt Cloud run fails"""

    pass


class DbtCloudRunCanceled(DbtCloudBaseException):
    """Raised when a dbt Cloud run has been canceled before completion"""

    pass


class DbtCloudRunTimedOut(DbtCloudBaseException):
    """Raised when a dbt Cloud run does not complete in the provided time"""

    pass


class DbtCloudListArtifactsFailed(DbtCloudBaseException):
    """Raised when dbt Cloud artifacts cannot be listed"""

    pass


def trigger_job_run(
    account_id: int,
    job_id: int,
    token: str,
    cause: str,
    domain: str,
    additional_args: dict,
) -> dict:
    """
    Trigger a dbt Cloud job run

    Args:
        - account_id (int): dbt Cloud account ID
        - job_id (int): dbt Cloud job ID
        - token (str): dbt Cloud token
        - cause (str): the reason describing why the job run is being triggered
        - domain (str): The domain the function should call (e.g. `cloud.getdbt.com`).
        - additional_args (dict): additional information to pass to the Trigger Job Run API

    Returns:
        - The trigger run result, namely the "data" key in the API response

    Raises:
        - TriggerDbtCloudRunFailed: when the response code is != 200
    """
    data = additional_args if additional_args else {}
    data["cause"] = cause
    trigger_request = requests.post(
        url=__DBT_CLOUD_TRIGGER_JOB_API_ENDPOINT_V2.format(
            accountId=account_id, jobId=job_id, apiDomain=domain
        ),
        headers={"Authorization": f"Bearer {token}", **USER_AGENT_HEADER},
        json=data,
    )

    if trigger_request.status_code != 200:
        raise TriggerDbtCloudRunFailed(trigger_request.reason)

    return trigger_request.json()["data"]


def wait_for_job_run(
    account_id: int,
    token: str,
    run_id: int,
    max_wait_time: int = None,
    domain: str = None,
) -> dict:
    """
    Get a dbt Cloud job run.
    Please note that this function will fail if any call to dbt Cloud APIs fail.

    Args:
        - account_id (int): dbt Cloud account ID
        - token (str): dbt Cloud token
        - run_id (int): dbt Cloud job run ID
        - max_wait_time (int): the number od seconds to wait for the job to complete
        - domain (str): The domain the function should call (e.g. `cloud.getdbt.com`).

    Returns:
        - The job run result, namely the "data" key in the API response

    Raises:
        - DbtCloudRunFailed: if "finished_at" is not None and the result status == 20
        - DbtCloudRunCanceled: if "finished_at" is not None and the result status == 30
        - DbtCloudRunTimedOut: if run does not finish before provided max_wait_time
    """
    wait_time_between_api_calls = 10
    elapsed_wait_time = 0
    while not max_wait_time or elapsed_wait_time <= max_wait_time:
        get_run_request = requests.get(
            url=__DBT_CLOUD_GET_RUN_API_ENDPOINT_V2.format(
                accountId=account_id, runId=run_id, apiDomain=domain
            ),
            headers={"Authorization": f"Bearer {token}", **USER_AGENT_HEADER},
        )

        if get_run_request.status_code != 200:
            raise GetDbtCloudRunFailed(get_run_request.reason)

        result = get_run_request.json()["data"]
        if result["finished_at"]:
            if result["status"] == 10:
                return result
            elif result["status"] == 20:
                raise DbtCloudRunFailed(f"Job run with ID: {run_id} failed.")
            elif result["status"] == 30:
                raise DbtCloudRunCanceled(f"Job run with ID: {run_id} cancelled.")
        sleep(wait_time_between_api_calls)
        elapsed_wait_time += wait_time_between_api_calls

    raise DbtCloudRunTimedOut(
        f"Max attempts reached while checking status of job run with ID: {run_id}"
    )


def list_run_artifact_links(
    account_id: int,
    run_id: int,
    token: str,
    domain: str = None,
) -> List[Tuple[str, str]]:
    """
    Lists URLs that can be used to download artifacts from a dbt run

    Args:
        - account_id (int): dbt Cloud account ID
        - run_id (int): dbt Cloud job run ID
        - token (str): dbt Cloud token
        - domain (str): The domain the function should call (e.g. `cloud.getdbt.com`).

    Returns:
        - List of artifact download URLs

    Raises:
        - DbtCloudListArtifactsFailed: if API to list dbt artifacts fails

    """

    list_run_artifact_response = requests.get(
        url=__DBT_CLOUD_LIST_RUN_ARTIFACTS_ENDPOINT_V2.format(
            accountId=account_id, runId=run_id, apiDomain=domain
        ),
        headers={"Authorization": f"Bearer {token}", **USER_AGENT_HEADER},
    )
    if list_run_artifact_response.status_code != 200:
        raise DbtCloudListArtifactsFailed(list_run_artifact_response.reason)

    artifact_paths = list_run_artifact_response.json().get("data")
    return [
        (
            __DBT_CLOUD_GET_RUN_ARTIFACT_ENDPOINT_V2.format(
                accountId=account_id, runId=run_id, path=artifact_path, apiDomain=domain
            ),
            artifact_path,
        )
        for artifact_path in artifact_paths
    ]



from pathlib import Path
from typing import Optional

import snowflake.connector as sf
from snowflake.connector.cursor import SnowflakeCursor

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class SnowflakeQuery(Task):
    """
    Task for executing a query against a Snowflake database.

    Args:
        - account (str): Snowflake account name, see Snowflake connector
             package documentation for details.
        - user (str): User name used to authenticate.
        - password (str, optional): Password used to authenticate.
            `password` or `private_key` must be present, unless `authenticator` is provided.
        - private_key (bytes, optional): PEM to authenticate.
            `password` or `private_key` must be present, unless `authenticator` is provided.
        - database (str, optional): Name of the default database to use.
        - schema (int, optional): Name of the default schema to use.
        - role (str, optional): Name of the default role to use.
        - warehouse (str, optional): Name of the default warehouse to use.
        - query (str, optional): Query to execute against database.
        - data (tuple, optional): Values to use in query, must be specified using placeholder
            in query string.
        - autocommit (bool, optional): Set to True to autocommit, defaults to None, which
            takes Snowflake AUTOCOMMIT parameter.
        - cursor_type (SnowflakeCursor, optional): Specify the type of database
            cursor to use for the query, defaults to SnowflakeCursor.
        - authenticator (str, optional): Type of authenticator to use for initiating
            connection (oauth, externalbrowser...), refer to Snowflake
            [Python Connector API](https://docs.snowflake.com/en/user-guide/python-connector-api.html#connect)
            documentation for details. Note that `externalbrowser` will only work in an environment
            where a browser is available, default to None.
        - token (str, optional): OAuth or JWT Token to provide when authenticator
            is set to oauth, default to None.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """  # noqa

    def __init__(
        self,
        account: str = None,
        user: str = None,
        password: Optional[str] = None,
        private_key: Optional[bytes] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        role: Optional[str] = None,
        warehouse: Optional[str] = None,
        query: str = None,
        data: Optional[tuple] = None,
        autocommit: Optional[bool] = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
        authenticator: Optional[str] = None,
        token: Optional[str] = None,
        **kwargs,
    ):
        self.account = account
        self.user = user
        self.password = password
        self.private_key = private_key
        self.database = database
        self.schema = schema
        self.role = role
        self.warehouse = warehouse
        self.query = query
        self.data = data
        self.autocommit = autocommit
        self.cursor_type = cursor_type
        self.authenticator = authenticator
        self.token = token
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "account",
        "user",
        "password",
        "private_key",
        "database",
        "schema",
        "role",
        "warehouse",
        "query",
        "data",
        "autocommit",
        "cursor_type",
        "authenticator",
        "token",
    )
    def run(
        self,
        account: str = None,
        user: str = None,
        password: Optional[str] = None,
        private_key: Optional[bytes] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        role: Optional[str] = None,
        warehouse: Optional[str] = None,
        query: str = None,
        data: Optional[tuple] = None,
        autocommit: Optional[bool] = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
        authenticator: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """
        Task run method. Executes a query against Snowflake database.

        Args:
            - account (str, optional): Snowflake account name, see Snowflake connector
                package documentation for details.
            - user (str, optional): User name used to authenticate.
            - password (str, optional): Password used to authenticate.
                `password` or `private_key` must be present, unless `authenticator` is provided
            - private_key (bytes, optional): PEM to authenticate.
                `password` or `private_key` must be present, unless `authenticator` is provided
            - database (str, optional): Name of the default database to use.
            - schema (int, optional): Name of the default schema to use.
            - role (str, optional): Name of the default role to use.
            - warehouse (str, optional): Name of the default warehouse to use.
            - query (str, optional): Query to execute against database.
            - data (tuple, optional): Values to use in query, must be specified using placeholder
                in query string.
            - autocommit (bool, optional): Set to True to autocommit, defaults to None, which
                takes Snowflake AUTOCOMMIT parameter.
            - cursor_type (SnowflakeCursor, optional): Specify the type of database
                cursor to use for the query, defaults to SnowflakeCursor.
            - authenticator (str, optional): Type of authenticator to use for initiating
                connection (oauth, externalbrowser...), refer to Snowflake
                [Python Connector API](https://docs.snowflake.com/en/user-guide/python-connector-api.html#connect)
                documentation for details. Note that `externalbrowser` will only work in an environment
                where a browser is available, default to None.
            - token (str, optional): OAuth or JWT Token to provide when authenticator
                is set to oauth, default to None.

        Returns:
            - List[List]: Output of cursor.fetchall().

        Raises:
            - ValueError: If a required parameter is not supplied.
            - DatabaseError: If exception occurs when executing the query.
        """  # noqa
        auth_kwargs = (password, private_key, authenticator, token)
        if not account:
            raise ValueError("An account must be provided")
        if all(auth_kwarg is None for auth_kwarg in auth_kwargs):
            raise ValueError(
                f"An authentication keyword must be provided: {auth_kwargs}"
            )
        if not query:
            raise ValueError("A query string must be provided")

        # build the connection parameter dictionary
        # we will remove `None` values next
        connect_params = {
            "account": account,
            "user": user,
            "password": password,
            "private_key": private_key,
            "database": database,
            "schema": schema,
            "role": role,
            "warehouse": warehouse,
            "autocommit": autocommit,
            "authenticator": authenticator,
            "token": token,
            # required to track task's usage in the Snowflake Partner Network Portal
            "application": f"Prefect_{self.__class__.__name__}",
        }

        # filter out unset values
        connect_params = {
            param: value for param, value in connect_params.items() if value is not None
        }

        # connect to database, open cursor
        conn = sf.connect(**connect_params)
        # context manager automatically rolls back failed transactions and closes
        with conn:
            with conn.cursor(cursor_type) as cursor:
                executed = cursor.execute(query, params=data).fetchall()
        return executed


class SnowflakeQueriesFromFile(Task):
    """
    Task for executing queries loaded from a file against a Snowflake database.
    Return a list containings the results of the queries.

    Note that using execute_string() is vulnerable to SQL injection.

    Args:
        - account (str, optional): Snowflake account name, see Snowflake connector
             package documentation for details.
        - user (str, optional): User name used to authenticate.
        - password (str, optional): Password used to authenticate.
            `password` or `private_key` must be present, unless `authenticator` is provided.
        - private_key (bytes, optional): PEM to authenticate.
            `password` or `private_key` must be present, unless `authenticator` is provided.
        - database (str, optional): Name of the default database to use.
        - schema (int, optional): Name of the default schema to use.
        - role (str, optional): Name of the default role to use.
        - warehouse (str, optional): Name of the default warehouse to use.
        - file_path (str, optional): File path to load query from.
        - autocommit (bool, optional): Set to True to autocommit, defaults to None, which
            takes Snowflake AUTOCOMMIT parameter.
        - cursor_type (SnowflakeCursor, optional): Specify the type of database
            cursor to use for the query, defaults to SnowflakeCursor.
        - authenticator (str, optional): Type of authenticator to use for initiating
            connection (oauth, externalbrowser...), refer to Snowflake
            [Python Connector API](https://docs.snowflake.com/en/user-guide/python-connector-api.html#connect)
            documentation for details. Note that `externalbrowser` will only work in an environment
            where a browser is available, default to None.
        - token (str, optional): OAuth or JWT Token to provide when authenticator
            is set to oauth, default to None.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """  # noqa

    def __init__(
        self,
        account: str = None,
        user: str = None,
        password: Optional[str] = None,
        private_key: Optional[bytes] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        role: Optional[str] = None,
        warehouse: Optional[str] = None,
        file_path: str = None,
        autocommit: Optional[bool] = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
        authenticator: Optional[str] = None,
        token: Optional[str] = None,
        **kwargs,
    ):
        self.account = account
        self.user = user
        self.password = password
        self.private_key = private_key
        self.database = database
        self.schema = schema
        self.role = role
        self.warehouse = warehouse
        self.file_path = file_path
        self.autocommit = autocommit
        self.cursor_type = cursor_type
        self.authenticator = authenticator
        self.token = token
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "account",
        "user",
        "password",
        "private_key",
        "database",
        "schema",
        "role",
        "warehouse",
        "file_path",
        "autocommit",
        "cursor_type",
        "authenticator",
        "token",
    )
    def run(
        self,
        account: str = None,
        user: str = None,
        password: Optional[str] = None,
        private_key: Optional[bytes] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        role: Optional[str] = None,
        warehouse: Optional[str] = None,
        file_path: str = None,
        autocommit: Optional[bool] = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
        authenticator: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """
        Task run method. Executes a query against Snowflake database.

        Args:
            - account (str): Snowflake account name, see Snowflake connector
                package documentation for details.
            - user (str): User name used to authenticate.
            - password (str, optional): Password used to authenticate.
                `password` or `private_key` must be present, unless `authenticator` is provided.
            - private_key (bytes, optional): PEM to authenticate.
                `password` or `private_key` must be present, unless `authenticator` is provided.
            - database (str, optional): Name of the default database to use.
            - schema (int, optional): Name of the default schema to use.
            - role (str, optional): Name of the default role to use.
            - warehouse (str, optional): Name of the default warehouse to use.
            - file_path (str, optional): File path to load query from.
            - autocommit (bool, optional): Set to True to autocommit, defaults to None, which
                takes Snowflake AUTOCOMMIT parameter.
            - cursor_type (SnowflakeCursor, optional): Specify the type of database
                cursor to use for the query, defaults to SnowflakeCursor.
            - authenticator (str, optional): Type of authenticator to use for initiating
                connection (oauth, externalbrowser...), refer to Snowflake
                [Python Connector API](https://docs.snowflake.com/en/user-guide/python-connector-api.html#connect)
                documentation for details. Note that `externalbrowser` will only work in an environment
                where a browser is available, default to None.
            - token (str, optional): OAuth or JWT Token to provide when authenticator
                is set to oauth, default to None.

        Returns:
            - List[List]: Containing the results of the different queries executed.

        Raises:
            - ValueError: If query parameter is None or a blank string.
            - DatabaseError: If exception occurs when executing the query.
            - FileNotFoundError: If File does not exist.
        """  # noqa
        auth_kwargs = (password, private_key, authenticator, token)
        if account is None:
            raise ValueError("An account must be provided")
        if user is None:
            raise ValueError("A user must be provided")
        if all(auth_kwarg is None for auth_kwarg in auth_kwargs):
            raise ValueError(
                f"An authentication keyword must be provided: {auth_kwargs}"
            )
        if file_path is None:
            raise ValueError("A file path must be provided")

        # build the connection parameter dictionary
        # we will remove `None` values next
        connect_params = {
            "account": account,
            "user": user,
            "password": password,
            "private_key": private_key,
            "database": database,
            "schema": schema,
            "role": role,
            "warehouse": warehouse,
            "autocommit": autocommit,
            "authenticator": authenticator,
            "token": token,
            # required to track task's usage in the Snowflake Partner Network Portal
            "application": f"Prefect_{self.__class__.__name__}",
        }

        # filter out unset values
        connect_params = {
            param: value for param, value in connect_params.items() if value is not None
        }

        # connect to database, open cursor
        conn = sf.connect(**connect_params)

        # try to execute query
        # context manager automatically rolls back failed transactions
        # load query from file
        query = Path(file_path).read_text()
        with conn:
            cursor_list = conn.execute_string(query, cursor_class=cursor_type)
            result = [cursor.fetchall() for cursor in cursor_list]
        return result



from requests import Session
import time
from typing import Dict

from prefect.engine.signals import FAIL


class HightouchClient:

    __HIGHTOUCH_BASE_URL = "https://api.hightouch.io"
    __HIGHTOUCH_START_NEW_SYNC_RUN_URL = f"{__HIGHTOUCH_BASE_URL}/api/v2/rest/run"
    __HIGHTOUCH_GET_SYNC_RUN_STATUS = f"{__HIGHTOUCH_BASE_URL}/api/v2/rest/sync"

    def __init__(self, api_key: str):
        """
        Create an `HightouchClient`.

        Args:
            - api_key (str): The API kye to use to authenticate API calls on Hightouch.
        """
        self.api_key = api_key
        self.session = Session()
        self.session.headers = {
            "Content-type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }

    def get_sync_run_status(self, sync_id: int) -> Dict:
        """
        Return the status of a sync run.
        The status is obtained by calling the
        [Get sync run status API](https://hightouch.io/docs/syncs/api/#get-the-status-of-a-sync-run).

        Args:
            - sync_id (int): The sync identifier.

        Raises:
            - `prefect.engine.signals.FAIL` if the response status code is not 200.

        Returns:
            - The JSON response containing information about the status of
                the sync run.
        """
        url = f"{self.__HIGHTOUCH_GET_SYNC_RUN_STATUS}/{sync_id}"
        with self.session.get(url) as response:
            if response.status_code != 200:
                msg = f"Error while retrieving sync run status. Error is: {response.reason}."
                raise FAIL(message=msg)

            return response.json()

    def start_sync_run(
        self,
        sync_id: int,
        wait_for_completion: bool,
        wait_time_between_api_calls: int,
        max_wait_time: int,
    ) -> Dict:
        """
        Start a new sync run.
        Optionally, wait for run completion.
        The sync run is triggered via the
        [Start new sync run API](https://hightouch.io/docs/syncs/api/#start-a-new-sync-run).

        Args:
            - sync_id (int): The sync identifier.
            - wait_for_completion (bool): Whether to wait for the sync run completion or not.
            - wait_time_between_api_calls (int): The number of seconds to wait between API calls.
                This is used only if `wait_for_completion` is `True`.
            - max_wait_time (int): The maximum number of seconds to wait for the sync run to complete.
                This is used only if `wait_for_completion` is `True`.

        Raises:
            - `prefect.engine.signals.FAIL` if the sync run takes more
                than `max_wait_time` seconds to complete.

        Returns:
            - If `wait_for_completion` is `True`, returns the JSON response
                containing the status of the sync run.
            - If `wait_for_completion` is `False`, returns the JSON response
                containing information about the start sync run action.
        """
        url = f"{self.__HIGHTOUCH_START_NEW_SYNC_RUN_URL}/{sync_id}"
        with self.session.post(url) as response:
            if response.status_code != 200:
                msg = f"Error while starting sync run. Error is: {response.reason}."
                raise FAIL(message=msg)

            start_sync_response = response.json()
            if wait_for_completion:
                elapsed_wait_time = 0
                sync_status = None
                while not max_wait_time or elapsed_wait_time <= max_wait_time:

                    sync_status_response = self.get_sync_run_status(sync_id=sync_id)
                    sync_status = sync_status_response["sync"]["sync_status"]

                    if sync_status == "success":
                        return sync_status_response
                    else:
                        time.sleep(wait_time_between_api_calls)
                        elapsed_wait_time += wait_time_between_api_calls

                msg = "Sync run exceeded `max_wait_time`"
                raise FAIL(message=msg)

            else:
                return start_sync_response



"""
This module contains a collection of tasks for interacting with MySQL databases via
the pymysql library.
"""
try:
    from prefect.tasks.mysql.mysql import MySQLExecute, MySQLFetch
except ImportError as import_error:
    raise ImportError(
        'Using `prefect.tasks.mysql` requires Prefect to be installed with the "mysql" extra.'
    ) from import_error

__all__ = ["MySQLExecute", "MySQLFetch"]



import os.path
import contextlib
from typing import Optional
from paramiko import Transport, SFTPClient, SFTPError

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class SftpDownload(Task):
    """
    Task for downloading files from an SFTP server.
    Downloads remote file into sftp_downloads/ folder by default.

    Args:
        - host (str): Name of the host to use.
        - username (str): Username used to authenticate.
        - password (str): Password used to authenticate.
        - port_number (int): The port to connect to the server.
        - remote_path (str): The remote SFTP file path.
        - local_path (str): The local file path to download file to.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the Task constructor.

    Raises:
        - ValueError: Raised if a required parameter is not supplied.
        - ClientError: Raised if exception occurs when connecting/downloading from the server.
    """

    def __init__(
        self,
        host: str = None,
        username: str = None,
        password: str = None,
        port_number: int = None,
        remote_path: str = None,
        local_path: str = None,
        **kwargs,
    ):
        self.host = host
        self.username = username
        self.password = password
        self.port_number = port_number
        self.remote_path = remote_path
        self.local_path = local_path
        super().__init__(**kwargs)

    def _create_connection(self) -> Optional[SFTPClient]:
        """
        Initialize the connection with the SFTP server.
        :return: Optional[SFTPClient]
        """
        transport = Transport(sock=(self.host, self.port_number))
        transport.connect(username=self.username, password=self.password)
        # self._connection = SFTPClient.from_transport(transport)
        connection = SFTPClient.from_transport(transport)
        self.logger.info(f"connected to {self.host}, {self.port_number}")
        return connection

    def file_exists(self, remote_path: str, conn: SFTPClient):
        """
        Checks if file exists in remote path or not.

        Args:
            - remote_path (str): Remote file path to check if file exists.
            - conn (SFTPClient): The connection object for the SFTP server.

        """
        try:
            self.logger.info(f"remote path : {remote_path}")
            conn.stat(remote_path)
        except SFTPError as e:
            self.logger.debug(
                f"The specified file on this '{remote_path}' remote_path does not exist."
            )
            raise e

    @defaults_from_attrs(
        "host", "username", "password", "port_number", "remote_path", "local_path"
    )
    def run(
        self,
        host: str = None,
        username: str = None,
        password: str = None,
        port_number: int = None,
        remote_path: str = None,
        local_path: str = None,
    ):
        """
        Task for downloading files from an SFTP server.

        Args:
            - host (str): Name of the host to use.
            - username (str): Username used to authenticate.
            - password (str): Password used to authenticate.
            - port_number (int): The port to connect to the server.
            - remote_path (str): The remote SFTP file path.
            - local_path (str): The local file path to download file to.

        Raises:
            - ValueError: Raised if a required parameter is not supplied.
            - ClientError: Raised if exception occurs when connecting/downloading from the server.
        """
        if not host:
            raise ValueError("A host name must be provided")
        if not username:
            raise ValueError("User name must be provided")
        if not password:
            raise ValueError("A password must be provided")
        if not port_number:
            raise ValueError("A port_number name must be provided")
        if not remote_path:
            raise ValueError("A remote_path must be provided")

        # set default to local path if arg not provided
        self.local_path = (
            "sftp_downloads/" + remote_path.split("/")[-1]
            if local_path is None
            else local_path
        )

        # check if local path exists or not
        local_dir = "/".join(self.local_path.split("/")[:-1]) + "/"
        if not os.path.isdir(local_dir):
            os.mkdir(local_dir)

        with contextlib.closing(self._create_connection()) as conn:
            self.file_exists(remote_path, conn)
            conn.get(remote_path, local_path, callback=None)


class SftpUpload(Task):
    """
    Task for uploading files to an SFTP server.

    Args:
        - host (str): Name of the host to use.
        - username (str): Username used to authenticate.
        - password (str): Password used to authenticate.
        - port_number (int): The port number to connect to the server.
        - remote_path (str): The remote SFTP file path.
        - local_path (str): The local file path to from upload.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the Task constructor.

    Raises:
        - ValueError: Raised if a required parameter is not supplied.
        - ClientError: Raised if exception occurs when connecting/uploading to the server.
    """

    def __init__(
        self,
        host: str = None,
        username: str = None,
        password: str = None,
        port_number: int = None,
        remote_path: str = None,
        local_path: str = None,
        **kwargs,
    ):
        self.host = host
        self.username = username
        self.password = password
        self.port_number = port_number
        self.remote_path = remote_path
        self.local_path = local_path
        super().__init__(**kwargs)

    def _create_connection(self) -> Optional[SFTPClient]:
        """
        Initialize the connection with the SFTP server.
        :return: Optional[SFTPClient]
        """
        transport = Transport(sock=(self.host, self.port_number))
        transport.connect(username=self.username, password=self.password)
        connection = SFTPClient.from_transport(transport)
        self.logger.info(f"connected to {self.host}, {self.port_number}")
        return connection

    @defaults_from_attrs(
        "host", "username", "password", "port_number", "remote_path", "local_path"
    )
    def run(
        self,
        host: str = None,
        username: str = None,
        password: str = None,
        port_number: int = None,
        remote_path: str = None,
        local_path: str = None,
    ):
        """
        Task for uploading files to an SFTP server.

        Args:
            - host (str): Name of the host to use.
            - username (str): Username used to authenticate.
            - password (str): Password used to authenticate.
            - port_number (int): The port number to connect to the server.
            - remote_path (str): The remote SFTP file path.
            - local_path (str): The local file path to upload from.

        Raises:
            - ValueError: Raised if a required parameter is not supplied.
            - ClientError: Raised if exception occurs when connecting/uploading from the server.
        """
        if not host:
            raise ValueError("A host name must be provided")
        if not username:
            raise ValueError("User name must be provided")
        if not password:
            raise ValueError("A password must be provided")
        if not port_number:
            raise ValueError("A port_number name must be provided")
        if not remote_path:
            raise ValueError("A remote_path must be provided")
        if not local_path:
            raise ValueError("A local_path must be provided")

        with contextlib.closing(self._create_connection()) as conn:
            conn.put(
                localpath=local_path,
                remotepath=remote_path,
                confirm=True,
            )



from typing import Dict

from prefect import Task
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class StepActivate(Task):
    """
    Task for activating an AWS Step Function.

    Args:
        - state_machine_arn (str): the Amazon Resource Name (ARN) of the state machine
            to execute
        - execution_name (str): the name of the execution, this name must be unique for
            your AWS account, region, and state machine for 90 days
        - execution_input (str, optional, DEPRECATED): string that contains the JSON input data for
            the execution, defaults to `"{}"`. Setting `execution_input` through the constructor is
            deprecated. Set it when running this task instead.
        - boto_kwargs (dict, optional): additional keyword arguments to forward to the boto client.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        state_machine_arn: str = None,
        execution_name: str = None,
        execution_input: str = "{}",
        boto_kwargs: dict = None,
        **kwargs,
    ):
        self.state_machine_arn = state_machine_arn
        self.execution_name = execution_name
        self.execution_input = execution_input

        if execution_name is not None:
            self.logger.warning(
                f"It's not recommended to set execution_name={execution_name} in the StepActivate "
                f"constructor, because the name has to be unique across executions in your account "
                f"and AWS region. Set execution_name when running your task instead."
            )

        if boto_kwargs is None:
            self.boto_kwargs = {}
        else:
            self.boto_kwargs = boto_kwargs

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "state_machine_arn",
        "execution_name",
        "execution_input",
        "boto_kwargs",
    )
    def run(
        self,
        credentials: dict = None,
        state_machine_arn: str = None,
        execution_name: str = None,
        execution_input: str = None,
        boto_kwargs: dict = None,
    ) -> Dict:
        """
        Task run method. Activates AWS Step function.

        Args:
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
            - state_machine_arn (str): the Amazon Resource Name (ARN) of the state machine
                to execute
            - execution_name (str): the name of the execution, this name must be unique for
                your AWS account, region, and state machine for 90 days
            - execution_input (str, optional): string that contains the JSON input data for
                the execution, defaults to `"{}"`
            - boto_kwargs (dict, optional): additional keyword arguments to forward to the
                boto client.

        Returns:
            - dict: response from AWS StartExecution endpoint
        """
        if not state_machine_arn:
            raise ValueError("A state machine ARN must be provided")

        step_client = get_boto_client(
            "stepfunctions", credentials=credentials, **boto_kwargs
        )

        response = step_client.start_execution(
            stateMachineArn=state_machine_arn,
            name=execution_name,
            input=execution_input,
        )

        return response



from typing import Union, Iterable

import yaml

from prefect.run_configs.base import RunConfig
from prefect.utilities.filesystems import parse_path


class ECSRun(RunConfig):
    """Configure a flow-run to run as an ECS Task.

    ECS Tasks are composed of task definitions and runtime parameters.

    Task definitions can be configured using either the `task_definition`,
    `task_definition_path`, or `task_definition_arn` parameters. If neither is
    specified, the default configured on the agent will be used. At runtime
    this task definition will be registered once per flow version - subsequent
    runs of the same flow version will reuse the existing definition.

    Runtime parameters can be specified via `run_task_kwargs`. These will be
    merged with any runtime parameters configured on the agent when starting
    the task.

    Note that certain Prefect specific environment variables defined within the
    `task_definition` will be overwritten by the ECS Agent when a new task is
    run (see method get_run_task_kwargs of the ECSAgent). For example,
    although the variable `PREFECT__LOGGING__LEVEL` might be defined within
    `containerDefinitions` (of the `task_definition`), it will first be
    ovewritten with the Prefect config and then by `env` (if defined). Therefore,
    do not set any Prefect specific environment variables within `task_definition`,
    `task_definition_path` or `task_definition_arn`. Instead, use the top-level
    `ECSRun.env` setting.

    Args:
        - task_definition (dict, optional): An in-memory task definition spec
            to use. The flow will be executed in a container named `flow` - if
            a container named `flow` isn't part of the task definition, Prefect
            will add a new container with that name.  See the
            [ECS.Client.register_task_definition][1] docs for more information
            on task definitions. Note that this definition will be stored
            directly in Prefect Cloud/Server - use `task_definition_path`
            instead if you wish to avoid this.
        - task_definition_path (str, optional): Path to a task definition spec
            to use. If you started your agent in a local process (with
            `prefect agent ecs start --label your-label`), then you can reference a path
            on that machine (use either no file scheme, or a `file`/`local` scheme).
            The task definition will then be loaded on initialization
            and stored on the `ECSRun` object as the `task_definition` field.
            Otherwise the task definition will be loaded at runtime on the
            agent.  Supported runtime file schemes include `s3`, `gcs`, and
            `agent` (for paths local to the runtime agent).
        - task_definition_arn (str, optional): A pre-registered task definition
            ARN to use (either `family:revision`, or a full task
            definition ARN). This task definition must include a container
            named `flow` (which will be used to run the flow).
        - image (str, optional): The image to use for this task. If not
            provided, will be either inferred from the flow's storage (if using
            `Docker` storage), or use the default configured on the agent. Note that
            when using an ECR image, you need to explicitly provide the
            `execution_role_arn`, unless you set this role explicitly in a custom `task_definition`.
            This means: either you provide a custom `task_definition` that contains
            both the image and `execution_role_arn` (with ECR permissions),
            or you provide both a custom `image` and an `execution_role_arn`
            simultaneously.
        - env (dict, optional): Additional environment variables to set on the task.
        - cpu (int or str, optional): The amount of CPU available to the task. Can
            be ether an integer in CPU units (e.g. `1024`), or a string using
            vCPUs (e.g. `"1 vcpu"`). Note that ECS imposes strict limits on
            what values this can take, see the [ECS documentation][2] for more
            information.
        - memory (int or str, optional): The amount of memory available to the
            task. Can be ether an integer in MiB (e.g. `1024`), or a string
            with units (e.g. `"1 GB"`). Note that ECS imposes strict limits on
            what values this can take, see the [ECS documentation][2] for more
            information.
        - task_role_arn (str, optional): The full ARN of the IAM role
            to use for this task. If you provide a custom `task_definition` that
            already contains the `task_role_arn`, then you can skip this argument.
            You can also skip it when your flow doesn't need access to any AWS
            resources such as S3. But if you use S3 storage (or results) and you don't set
            a custom `task_definition`, this role must be set explicitly.
        - execution_role_arn (str, optional): The execution role ARN to use
            when registering a task definition for this task. If you provide a custom
            `task_definition` that already contains the `execution_role_arn`, then you
            can skip this argument. But if you don't set any custom `task_definition`,
            and you supply an ECR `image`, then you need to set an `execution_role_arn`
            explicitly to grant ECR access.
        - run_task_kwargs (dict, optional): Additional keyword arguments to
            pass to `run_task` when starting this task. It should be used only for
            runtime-specific arguments such as `cpu`, `memory`, `cluster`, or `launchType`,
            rather than `task_definition`-specific arguments.
            See the [ECS.Client.run_task][3] docs for more information.
        - labels (Iterable[str], optional): An iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work

    Examples:

    1) Use the defaults set on the agent (assuming label "prod"):

    ```python
    flow.run_config = ECSRun(labels=["prod"])
    ```

    2) Use a custom task definition uploaded to S3 as a YAML file. This task definition contains
    a container named "flow", as well as a custom execution role and task role. The flow should
    be picked up for execution by an agent with label "prod" and should be deployed to a cluster
    called "prefectEcsCluster".

    ```python
    flow.run_config = ECSRun(
        labels=["prod"],
        task_definition_path="s3://bucket/flow_task_definition.yaml",
        run_task_kwargs=dict(cluster="prefectEcsCluster"),
    )
    ```

    An example of `flow_task_definition.yaml`:
    ```yaml
    family: prefectFlow
    requiresCompatibilities:
        - FARGATE
    networkMode: awsvpc
    cpu: 1024
    memory: 2048
    taskRoleArn: arn:aws:iam::XXX:role/prefectTaskRole
    executionRoleArn: arn:aws:iam::XXX:role/prefectECSAgentTaskExecutionRole
    containerDefinitions:
    - name: flow
        image: "XXX.dkr.ecr.us-east-1.amazonaws.com/image_name:latest"
        essential: true
        environment:
        - name: AWS_RETRY_MODE
            value: "adaptive"
        - name: AWS_MAX_ATTEMPTS
            value: "10"
        logConfiguration:
        logDriver: awslogs
        options:
            awslogs-group: "/ecs/prefectEcsAgent"
            awslogs-region: "us-east-1"
            awslogs-stream-prefix: "ecs"
            awslogs-create-group: "true"
    ```

    3) Use a custom pre-registered task definition (referenced by `family:revision`
    or by a full ARN), then deploy the flow to a cluster called "prefectEcsCluster":

    ```python
    flow.run_config = ECSRun(
        labels=["prod"],
        task_definition_arn="prefectFlow:1",
        # or using a full ARN:
        # task_definition_arn="arn:aws:ecs:us-east-1:XXX:task-definition/prefectFlow:1"
        run_task_kwargs=dict(cluster="prefectEcsCluster"),
    )
    ```

    4) Let Prefect register a new task definition for a flow with S3 storage and a custom
    ECR image:

    ```python
    flow.run_config = ECSRun(
        labels=["prod"],
        task_role_arn="arn:aws:iam::XXXX:role/prefectTaskRole",
        execution_role_arn="arn:aws:iam::XXXX:role/prefectECSAgentTaskExecutionRole",
        image="XXXX.dkr.ecr.us-east-1.amazonaws.com/image_name:latest",
        run_task_kwargs=dict(cluster="prefectEcsCluster"),
    )
    ```

    5) Use the default task definition, but override the image and CPU:

    ```python
    flow.run_config = ECSRun(
        execution_role_arn="arn:aws:iam::XXXX:role/prefectECSAgentTaskExecutionRole",
        image="XXXX.dkr.ecr.us-east-1.amazonaws.com/image_name:latest",
        cpu="2 vcpu",
    )
    ```

    6) Use an explicit task definition stored in S3 as a YAML file,
    but override the image and CPU:

    ```python
    flow.run_config = ECSRun(
        task_definition="s3://bucket/flow_task_definition.yaml",
        image="XXXX.dkr.ecr.us-east-1.amazonaws.com/image_name:latest",
        cpu="2 vcpu",
    )
    ```

    7) An example flow showing ECSRun with a `task_definition` defined as dictionary,
    and S3 storage:
    ```python
    import prefect
    from prefect.storage import S3
    from prefect.run_configs import ECSRun
    from prefect import task, Flow
    from prefect.client.secrets import Secret


    FLOW_NAME = "ecs_demo_ecr"
    ACCOUNT_ID = Secret("AWS_ACCOUNT_ID").get()
    STORAGE = S3(
        bucket="your_bucket_name",
        key=f"flows/{FLOW_NAME}.py",
        stored_as_script=True,
        local_script_path=f"{FLOW_NAME}.py",
    )
    RUN_CONFIG = ECSRun(
        labels=["prod"],
        task_definition=dict(
            family=FLOW_NAME,
            requiresCompatibilities=["FARGATE"],
            networkMode="awsvpc",
            cpu=1024,
            memory=2048,
            taskRoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/prefectTaskRole",
            executionRoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/prefectECSAgentTaskExecutionRole",
            containerDefinitions=[
                dict(
                    name="flow",
                    image=f"{ACCOUNT_ID}.dkr.ecr.us-east-1.amazonaws.com/your_image_name:latest",
                )
            ],
        ),
        run_task_kwargs=dict(cluster="prefectEcsCluster"),
    )


    @task
    def say_hi():
        logger = prefect.context.get("logger")
        logger.info("Hi from Prefect %s from flow %s", prefect.__version__, FLOW_NAME)


    with Flow(FLOW_NAME, storage=STORAGE, run_config=RUN_CONFIG,) as flow:
        say_hi()

    ```

    [1]: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/\
ecs.html#ECS.Client.register_task_definition

    [2]: https://docs.aws.amazon.com/AmazonECS/latest/userguide/task-cpu-memory-error.html

    [3]: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/\
ecs.html#ECS.Client.run_task
    """

    def __init__(
        self,
        *,
        task_definition: dict = None,
        task_definition_path: str = None,
        task_definition_arn: str = None,
        image: str = None,
        env: dict = None,
        cpu: Union[int, str] = None,
        memory: Union[int, str] = None,
        task_role_arn: str = None,
        execution_role_arn: str = None,
        run_task_kwargs: dict = None,
        labels: Iterable[str] = None,
    ) -> None:
        super().__init__(env=env, labels=labels)

        if (
            sum(
                [
                    task_definition is not None,
                    task_definition_path is not None,
                    task_definition_arn is not None,
                ]
            )
            > 1
        ):
            raise ValueError(
                "Can only provide one of `task_definition`, `task_definition_path`, "
                "or `task_definition_arn`"
            )
        if task_definition_arn is not None and image is not None:
            raise ValueError(
                "Cannot provide both `task_definition_arn` and `image`, since an ECS "
                "task's `image` can only be configured as part of the task definition"
            )

        if task_definition_path is not None:
            parsed = parse_path(task_definition_path)
            if parsed.scheme == "file":
                with open(parsed.path) as f:
                    task_definition = yaml.safe_load(f)
                    task_definition_path = None

        if cpu is not None:
            cpu = str(cpu)
        if memory is not None:
            memory = str(memory)

        self.task_definition = task_definition
        self.task_definition_path = task_definition_path
        self.task_definition_arn = task_definition_arn
        self.image = image
        self.cpu = cpu
        self.memory = memory
        self.task_role_arn = task_role_arn
        self.execution_role_arn = execution_role_arn
        self.run_task_kwargs = run_task_kwargs



import pendulum
import requests
from typing import Any, Dict, List

from prefect.utilities.logging import get_logger
from prefect.utilities.graphql import format_graphql_request_error
from prefect.exceptions import ClientError


class MonteCarloClient:
    def __init__(
        self,
        api_key_id: str,
        api_token: str,
    ) -> None:
        self.api_key_id = api_key_id
        self.api_token = api_token
        self.logger = get_logger()
        self._api_url = "https://api.getmontecarlo.com/graphql"

    def _send_graphql_request(
        self, query: str, variables: dict = None
    ) -> Dict[str, Any]:
        response = requests.post(
            url=self._api_url,
            json=dict(query=query, variables=variables),
            headers={
                "x-mcd-id": self.api_key_id,
                "x-mcd-token": self.api_token,
                "Content-Type": "application/json",
            },
        )
        # Check if request returned a successful status
        try:
            response.raise_for_status()
        except requests.HTTPError as exc:
            if response.status_code == 400:
                # Create a custom-formatted err message for graphql errors which always
                # return a 400 status code and have "query" in the parameter dict
                try:
                    graphql_msg = format_graphql_request_error(response)
                except Exception:
                    # Fallback to a general message
                    graphql_msg = (
                        "This is likely caused by a poorly formatted GraphQL query or "
                        "mutation but the response could not be parsed for more details"
                    )
                raise ClientError(f"{exc}\n{graphql_msg}") from exc

            # Server-side and non-graphql errors will be raised without modification
            raise
        response = response.json()
        self.logger.debug(
            "Response: %s for request %s with variables %s", response, query, variables
        )
        return response

    def get_resources(self) -> List[Dict[str, Any]]:
        response = self._send_graphql_request(
            query="""
                    query {
                        getResources {
                            name
                            type
                            id
                            uuid
                            isDefault
                            isUserProvided
                        }
                    }
                    """
        )
        return response["data"]["getResources"]

    def create_or_update_tags_for_mcon(
        self, key: str, value: str, mcon: str
    ) -> Dict[str, Any]:
        response = self._send_graphql_request(
            query="""
                mutation($mcon_id: String!, $key: String!, $value: String!) {
                    createOrUpdateObjectProperty(mconId: $mcon_id,
                        propertyName: $key, propertyValue: $value) {
                        objectProperty {
                            id
                        }
                    }
                }
                """,
            variables=dict(mcon_id=mcon, key=key, value=value),
        )
        return response["data"]["createOrUpdateObjectProperty"]["objectProperty"]["id"]

    def create_or_update_lineage_node(
        self,
        node_name: str,
        object_id: str,
        object_type: str,
        resource_name: str,
    ):
        response = self._send_graphql_request(
            query="""
            mutation($node_name: String!, $object_id: String!, $object_type: String!,
            $resource_name: String! ) {
              createOrUpdateLineageNode(
                name: $node_name,
                objectId: $object_id,
                objectType: $object_type,
                resourceName: $resource_name,
              ){
                node{
                  nodeId
                  mcon
                }
              }
            }
            """,
            variables=dict(
                node_name=node_name,
                object_id=object_id,
                object_type=object_type,
                resource_name=resource_name,
            ),
        )
        return response["data"]["createOrUpdateLineageNode"]["node"]["mcon"]

    def create_or_update_lineage_node_with_multiple_tags(
        self,
        node_name: str,
        object_id: str,
        object_type: str,
        resource_name: str,
        tags: List[Dict[str, str]],
    ) -> str:
        response = self._send_graphql_request(
            query="""
                mutation($node_name: String!, $object_id: String!, $object_type: String!,
                $resource_name: String!, $tags: [ObjectPropertyInput]
                ) {
                  createOrUpdateLineageNode(
                    name: $node_name,
                    objectId: $object_id,
                    objectType: $object_type,
                    resourceName: $resource_name,
                    properties: $tags
                  ){
                    node{
                      mcon
                    }
                  }
                }
                """,
            variables=dict(
                node_name=node_name,
                object_id=object_id,
                object_type=object_type,
                resource_name=resource_name,
                tags=tags,
            ),
        )
        return response["data"]["createOrUpdateLineageNode"]["node"]["mcon"]

    def create_or_update_resource(self, resource_name: str, resource_type: str):
        response = self._send_graphql_request(
            query="""
            mutation($resource_name: String!, $resource_type: String!) {
              createOrUpdateResource(
                isDefault: false,
                name: $resource_name,
                type: $resource_type,
              ) {
                resource {
                  uuid
                }
              }
            }
            """,
            variables=dict(resource_name=resource_name, resource_type=resource_type),
        )
        return response["data"]["createOrUpdateResource"]["resource"]["uuid"]

    def create_or_update_lineage_edge(
        self, source: dict, destination: dict, expire_at: str = None
    ):
        if expire_at is None:
            expire_at = pendulum.now().add(days=1).isoformat()

        response = self._send_graphql_request(
            query="""
                mutation($destination_object_id: String!,
                $destination_object_type: String!,
                $destination_resource_name: String!,
                $source_object_id: String!, $source_object_type: String!,
                 $source_resource_name: String!, $expire_at: DateTime) {
                  createOrUpdateLineageEdge(
                    destination: {
                      objectId: $destination_object_id
                      objectType: $destination_object_type
                      resourceName: $destination_resource_name
                    }
                    source: {
                      objectId: $source_object_id
                      objectType: $source_object_type
                      resourceName: $source_resource_name
                    }
                    expireAt: $expire_at
                  ){
                    edge{
                      edgeId
                    }
                  }
                }
                """,
            variables=dict(
                destination_object_id=destination["object_id"],
                destination_object_type=destination["object_type"],
                destination_resource_name=destination["resource_name"],
                source_object_id=source["object_id"],
                source_object_type=source["object_type"],
                source_resource_name=source["resource_name"],
                expire_at=expire_at,
            ),
        )
        return response["data"]["createOrUpdateLineageEdge"]["edge"]["edgeId"]



import os
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
import json
from typing import Dict, List, Union

from .cubejs_utils import CubeJSClient


class CubeJSQueryTask(Task):

    """
    This task calls Cueb.js load API and returns the result
    as a JSON object.
    More info about Cube.js load API at
    https://cube.dev/docs/rest-api#api-reference-v-1-load.

    Args:
        - subdomain (str, optional): The subdomain to use to get the data.
            If provided, `subdomain` takes precedence over `url`.
            This is likely to be useful to Cube Cloud users.
        - url (str, optional): The URL to use to get the data.
            This is likely to be useful to users of self-hosted Cube.js.
        - api_secret (str, optional): The API secret used to generate an
            API token for authentication.
            If provided, it takes precedence over `api_secret_env_var`.
        - api_secret_env_var (str, optional): The name of the env var that contains
            the API secret to use to generate an API token for authentication.
            Defaults to `CUBEJS_API_SECRET`.
        - query (dict, list, optional): `dict` or `list` representing
            valid Cube.js queries.
            If you pass multiple queries, then be aware of Cube.js Data Blending.
            More info at https://cube.dev/docs/rest-api#api-reference-v-1-load
            and at https://cube.dev/docs/schema/advanced/data-blending.
            Query format can be found at: https://cube.dev/docs/query-format.
        - include_generated_sql: (bool, optional): Whether the return object should
            include SQL info or not.
            Default to `False`.
        - security_context (str, dict, optional): The security context to use
            during authentication.
            If the security context does not contain an expiration period,
            then a 7-day expiration period is added automatically.
            More info at: https://cube.dev/docs/security/context.
        - wait_time_between_api_calls (int, optional): The number of seconds to
            wait between API calls.
            Default to 10.
        - max_wait_time (int, optional): The number of seconds to wait for the
            Cube.js load API to return a response.
        - **kwargs (optional): Additional keyword arguments to pass to the
            standard Task initalization.

    """

    __CUBEJS_CLOUD_BASE_URL = "https://{subdomain}.cubecloud.dev"

    def __init__(
        self,
        subdomain: str = None,
        url: str = None,
        api_secret: str = None,
        api_secret_env_var: str = "CUBEJS_API_SECRET",
        query: Union[Dict, List[Dict]] = None,
        include_generated_sql: bool = False,
        security_context: Union[str, Dict] = None,
        wait_time_between_api_calls: int = 10,
        max_wait_time: int = None,
        **kwargs,
    ):
        self.subdomain = subdomain
        self.url = url
        self.api_secret = api_secret
        self.api_secret_env_var = api_secret_env_var
        self.query = query
        self.include_generated_sql = include_generated_sql
        self.security_context = security_context
        self.wait_time_between_api_calls = wait_time_between_api_calls
        self.max_wait_time = max_wait_time
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "subdomain",
        "url",
        "api_secret",
        "api_secret_env_var",
        "query",
        "include_generated_sql",
        "security_context",
        "wait_time_between_api_calls",
        "max_wait_time",
    )
    def run(
        self,
        subdomain: str = None,
        url: str = None,
        api_secret: str = None,
        api_secret_env_var: str = "CUBEJS_API_SECRET",
        query: Union[Dict, List[Dict]] = None,
        include_generated_sql: bool = False,
        security_context: Union[str, Dict] = None,
        wait_time_between_api_calls: int = 10,
        max_wait_time: int = None,
    ):
        """
        Task run method to perform a query using Cube.js load API.

        Args:
            - subdomain (str, optional): The subdomain to use to get the data.
                If provided, `subdomain` takes precedence over `url`.
                This is likely to be useful to Cube Cloud users.
            - url (str, optional): The URL to use to get the data.
                This is likely the preferred method for self-hosted Cube
                deployments.
                For Cube Cloud deployments, the URL should be in the form
                `https://<cubejs-generated-host>/cubejs-api`.
            - api_secret (str, optional): The API secret used to generate an
                API token for authentication.
                If provided, it takes precedence over `api_secret_env_var`.
            - api_secret_env_var (str, optional): The name of the env var that contains
                the API secret to use to generate an API token for authentication.
                Defaults to `CUBEJS_API_SECRET`.
            - query (dict, list, optional): `dict` or `list` representing
                valid Cube.js queries.
                If you pass multiple queries, then be aware of Cube.js Data Blending.
                More info at https://cube.dev/docs/rest-api#api-reference-v-1-load
                and at https://cube.dev/docs/schema/advanced/data-blending.
                Query format can be found at: https://cube.dev/docs/query-format.
            - include_generated_sql: (bool, optional): Whether the return object should
                include SQL info or not.
                Default to `False`.
            - security_context (str, dict, optional): The security context to use
                during authentication.
                If the security context does not contain an expiration period,
                then a 7-day expiration period is added automatically.
                More info at https://cube.dev/docs/security/context.
            - wait_time_between_api_calls (int, optional): The number of seconds to
                wait between API calls.
                Default to 10.
            - max_wait_time (int, optional): The number of seconds to wait for the
                Cube.js load API to return a response.

        Raises:
            - ValueError if both `subdomain` and `url` are missing.
            - ValueError if `api_token` is missing and `api_token_env_var` cannot be found.
            - ValueError if `query` is missing.
            - `prefect.engine.signals.FAIL` if the Cube.js load API fails.
            - `prefect.engine.signals.FAIL` if the Cube.js load API takes more than
                `max_wait_time` seconds to respond.

        Returns:
            - The Cube.js JSON response, augmented with SQL
                information if `include_generated_sql` is `True`.

        """

        if not subdomain and not url:
            raise ValueError("Missing both `subdomain` and `url`.")

        if not api_secret and api_secret_env_var not in os.environ:
            raise ValueError("Missing `api_secret` and `api_secret_env_var` not found.")

        if not query:
            raise ValueError("Missing `query`.")

        secret = api_secret if api_secret else os.environ[api_secret_env_var]

        wait_api_call_secs = (
            wait_time_between_api_calls if wait_time_between_api_calls > 0 else 10
        )

        cubejs_client = CubeJSClient(
            subdomain=subdomain,
            url=url,
            security_context=security_context,
            secret=secret,
            wait_api_call_secs=wait_api_call_secs,
            max_wait_time=max_wait_time,
        )

        params = {"query": json.dumps(query)}

        # Retrieve data from Cube.js
        data = cubejs_client.get_data(
            params=params, include_generated_sql=include_generated_sql
        )

        return data



from typing import Any, cast, Callable

from kubernetes import client
from kubernetes.stream import stream
from kubernetes.watch import Watch
from kubernetes.client.rest import ApiException

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
from prefect.utilities.kubernetes import get_kubernetes_client


class CreateNamespacedPod(Task):
    """
    Task for creating a namespaced pod on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    The arguments `body` and `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `body = {"info": "here"}` at
    instantiation and then provide `body = {"more": "info"}` at run time which will make
    `body = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - body (dict, optional): A dictionary representation of a Kubernetes V1Pod
            specification
        - namespace (str, optional): The Kubernetes namespace to create this pod in,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        body: dict = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.body = body or {}
        self.namespace = namespace
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "body", "namespace", "kube_kwargs", "kubernetes_api_key_secret"
    )
    def run(
        self,
        body: dict = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
    ) -> None:
        """
        Task run method.

        Args:
            - body (dict, optional): A dictionary representation of a Kubernetes V1Pod
                specification
            - namespace (str, optional): The Kubernetes namespace to create this pod in,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format

        Raises:
            - ValueError: if `body` is `None`
        """
        if not body:
            raise ValueError("A dictionary representing a V1Pod must be provided.")

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        body = {**self.body, **(body or {})}
        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}

        api_client.create_namespaced_pod(namespace=namespace, body=body, **kube_kwargs)


class DeleteNamespacedPod(Task):
    """
    Task for deleting a namespaced pod on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    The argument `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `kube_kwargs = {"info": "here"}` at
    instantiation and then provide `kube_kwargs = {"more": "info"}` at run time which will make
    `kube_kwargs = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - pod_name (str, optional): The name of a pod to delete
        - namespace (str, optional): The Kubernetes namespace to delete this pod from,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        pod_name: str = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.pod_name = pod_name
        self.namespace = namespace
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pod_name", "namespace", "kube_kwargs", "kubernetes_api_key_secret"
    )
    def run(
        self,
        pod_name: str = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        delete_option_kwargs: dict = None,
    ) -> None:
        """
        Task run method.

        Args:
            - pod_name (str, optional): The name of a pod to delete
            - namespace (str, optional): The Kubernetes namespace to delete this pod in,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format
            - delete_option_kwargs (dict, optional): Optional keyword arguments to pass to
                the V1DeleteOptions object (e.g. {"propagation_policy": "...",
                "grace_period_seconds": "..."}.

        Raises:
            - ValueError: if `pod_name` is `None`
        """
        if not pod_name:
            raise ValueError("The name of a Kubernetes pod must be provided.")

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}
        delete_option_kwargs = delete_option_kwargs or {}

        api_client.delete_namespaced_pod(
            name=pod_name,
            namespace=namespace,
            body=client.V1DeleteOptions(**delete_option_kwargs),
            **kube_kwargs
        )


class ListNamespacedPod(Task):
    """
    Task for listing namespaced pods on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    The argument `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `kube_kwargs = {"info": "here"}` at
    instantiation and then provide `kube_kwargs = {"more": "info"}` at run time which will make
    `kube_kwargs = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - namespace (str, optional): The Kubernetes namespace to list pods from,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"field_selector": "...", "label_selector": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.namespace = namespace
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs("namespace", "kube_kwargs", "kubernetes_api_key_secret")
    def run(
        self,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
    ) -> None:
        """
        Task run method.

        Args:
            - namespace (str, optional): The Kubernetes namespace to list pods from,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"field_selector": "...", "label_selector": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format

        Returns:
            - V1PodList: a Kubernetes V1PodList of the pods which are found
        """
        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}

        return api_client.list_namespaced_pod(namespace=namespace, **kube_kwargs)


class PatchNamespacedPod(Task):
    """
    Task for patching a namespaced pod on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    The arguments `body` and `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `body = {"info": "here"}` at
    instantiation and then provide `body = {"more": "info"}` at run time which will make
    `body = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - pod_name (str, optional): The name of a pod to patch
        - body (dict, optional): A dictionary representation of a Kubernetes V1Pod
            patch specification
        - namespace (str, optional): The Kubernetes namespace to patch this pod in,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        pod_name: str = None,
        body: dict = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.pod_name = pod_name
        self.body = body or {}
        self.namespace = namespace
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pod_name", "body", "namespace", "kube_kwargs", "kubernetes_api_key_secret"
    )
    def run(
        self,
        pod_name: str = None,
        body: dict = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
    ) -> None:
        """
        Task run method.

        Args:
            - pod_name (str, optional): The name of a pod to patch
            - body (dict, optional): A dictionary representation of a Kubernetes V1Pod
                patch specification
            - namespace (str, optional): The Kubernetes namespace to patch this pod in,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format

        Raises:
            - ValueError: if `body` is `None`
            - ValueError: if `pod_name` is `None`
        """
        if not body:
            raise ValueError(
                "A dictionary representing a V1Pod patch must be provided."
            )

        if not pod_name:
            raise ValueError("The name of a Kubernetes pod must be provided.")

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        body = {**self.body, **(body or {})}
        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}

        api_client.patch_namespaced_pod(
            name=pod_name, namespace=namespace, body=body, **kube_kwargs
        )


class ReadNamespacedPod(Task):
    """
    Task for reading a namespaced pod on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    The argument `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `kube_kwargs = {"info": "here"}` at
    instantiation and then provide `kube_kwargs = {"more": "info"}` at run time which will make
    `kube_kwargs = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - pod_name (str, optional): The name of a pod to read
        - namespace (str, optional): The Kubernetes namespace to read this pod from,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "exact": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        pod_name: str = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.pod_name = pod_name
        self.namespace = namespace
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pod_name", "namespace", "kube_kwargs", "kubernetes_api_key_secret"
    )
    def run(
        self,
        pod_name: str = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
    ) -> None:
        """
        Task run method.

        Args:
            - pod_name (str, optional): The name of a pod to read
            - namespace (str, optional): The Kubernetes namespace to read this pod in,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"pretty": "...", "exact": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format

        Returns:
            - V1Pod: a Kubernetes V1Pod matching the pod that was found

        Raises:
            - ValueError: if `pod_name` is `None`
        """
        if not pod_name:
            raise ValueError("The name of a Kubernetes pod must be provided.")

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}

        return api_client.read_namespaced_pod(
            name=pod_name, namespace=namespace, **kube_kwargs
        )


class ReplaceNamespacedPod(Task):
    """
    Task for replacing a namespaced pod on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    The arguments `body` and `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `body = {"info": "here"}` at
    instantiation and then provide `body = {"more": "info"}` at run time which will make
    `body = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - pod_name (str, optional): The name of a pod to replace
        - body (dict, optional): A dictionary representation of a Kubernetes V1Pod
            specification
        - namespace (str, optional): The Kubernetes namespace to patch this pod in,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        pod_name: str = None,
        body: dict = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.pod_name = pod_name
        self.body = body or {}
        self.namespace = namespace
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pod_name", "body", "namespace", "kube_kwargs", "kubernetes_api_key_secret"
    )
    def run(
        self,
        pod_name: str = None,
        body: dict = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
    ) -> None:
        """
        Task run method.

        Args:
            - pod_name (str, optional): The name of a pod to replace
            - body (dict, optional): A dictionary representation of a Kubernetes V1Pod
                specification
            - namespace (str, optional): The Kubernetes namespace to patch this pod in,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format

        Raises:
            - ValueError: if `body` is `None`
            - ValueError: if `pod_name` is `None`
        """
        if not body:
            raise ValueError("A dictionary representing a V1Pod must be provided.")

        if not pod_name:
            raise ValueError("The name of a Kubernetes pod must be provided.")

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        body = {**self.body, **(body or {})}
        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}

        api_client.replace_namespaced_pod(
            name=pod_name, namespace=namespace, body=body, **kube_kwargs
        )


class ReadNamespacedPodLogs(Task):
    """
    Task for reading logs from a namespaced pod on Kubernetes. Logs can be streamed by
    providing a `on_log_entry` function which then will be called for each log line. If
    `on_log_entry` = `None`, the task returns all logs for the pod until that point.

    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    Args:
        - pod_name (str, optional): The name of a pod to replace
        - namespace (str, optional): The Kubernetes namespace to patch this pod in,
            defaults to the `default` namespace
        - on_log_entry (Callable, optional): If provided, will stream the pod logs
            calling the callback for every line (and the task returns `None`). If not
            provided, the current pod logs will be returned immediately from the task.
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - container (str, optional): The name of the container to read logs from
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        pod_name: str = None,
        namespace: str = "default",
        on_log_entry: Callable = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        container: str = None,
        **kwargs: Any
    ):
        self.pod_name = pod_name
        self.namespace = namespace
        self.on_log_entry = on_log_entry
        self.kubernetes_api_key_secret = kubernetes_api_key_secret
        self.container = container

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pod_name",
        "namespace",
        "on_log_entry",
        "kubernetes_api_key_secret",
        "container",
    )
    def run(
        self,
        pod_name: str = None,
        namespace: str = "default",
        on_log_entry: Callable = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        container: str = None,
    ) -> None:
        """
        Task run method.

        Args:
            - pod_name (str, optional): The name of a pod to replace
            - namespace (str, optional): The Kubernetes namespace to patch this pod in,
                defaults to the `default` namespace
            - on_log_entry (Callable, optional): If provided, will stream the pod logs
                calling the callback for every line (and the task returns `None`). If not
                provided, the current pod logs will be returned immediately from the task.
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format
            - container (str, optional): The name of the container to read logs from

        Raises:
            - ValueError: if `pod_name` is `None`
        """
        if not pod_name:
            raise ValueError("The name of a Kubernetes pod must be provided.")

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        if on_log_entry is None:
            return api_client.read_namespaced_pod_log(
                name=pod_name, namespace=namespace, container=container
            )

        # From the kubernetes.watch documentation:
        # Note that watching an API resource can expire. The method tries to
        # resume automatically once from the last result, but if that last result
        # is too old as well, an `ApiException` exception will be thrown with
        # ``code`` 410.
        while True:
            try:
                stream = Watch().stream(
                    api_client.read_namespaced_pod_log,
                    name=pod_name,
                    namespace=namespace,
                    container=container,
                )

                for log in stream:
                    on_log_entry(log)

                return
            except ApiException as exception:
                if exception.status != 410:
                    raise


class ConnectGetNamespacedPodExec(Task):
    """
    Task for running a command in a namespaced pod on Kubernetes.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.
    This task will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file
    The argument `kube_kwargs` will perform an in-place update when the task
    is run. This means that it is possible to provide `kube_kwargs = {"info": "here"}` at
    instantiation and then provide `kube_kwargs = {"more": "info"}` at run time which will make
    `kube_kwargs = {"info": "here", "more": "info"}`. *Note*: Keys present in both instantiation
    and runtime will be replaced with the runtime value.

    Args:
        - pod_name (str, optional): The name of a pod in which the command is to be run
        - container_name (str, optional): The name of a container to use in the pod
        - exec_command (list, optional): the command to run in pod_name
        - namespace (str, optional): The Kubernetes namespace of the pod,
            defaults to the `default` namespace
        - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "exact": "..."}`)
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task
            constructor
    """

    def __init__(
        self,
        pod_name: str = None,
        container_name: str = None,
        exec_command: list = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
        **kwargs: Any
    ):
        self.pod_name = pod_name
        self.namespace = namespace
        self.container_name = container_name
        self.exec_command = exec_command
        self.kube_kwargs = kube_kwargs or {}
        self.kubernetes_api_key_secret = kubernetes_api_key_secret

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "pod_name",
        "container_name",
        "namespace",
        "kube_kwargs",
        "kubernetes_api_key_secret",
        "exec_command",
    )
    def run(
        self,
        pod_name: str = None,
        container_name: str = None,
        exec_command: list = None,
        namespace: str = "default",
        kube_kwargs: dict = None,
        kubernetes_api_key_secret: str = "KUBERNETES_API_KEY",
    ) -> None:
        """
        Task run method.

        Args:
            - pod_name (str, optional): The name of a pod in which the command is to be run
            - container_name (str, optional): The name of a container to use in the pod
            - exec_command (list, optional): the command to run in pod_name
            - namespace (str, optional): The Kubernetes namespace of the pod,
                defaults to the `default` namespace
            - kube_kwargs (dict, optional): Optional extra keyword arguments to pass to the
                Kubernetes API (e.g. `{"pretty": "...", "exact": "..."}`)
            - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
                which stored your Kubernetes API Key; this Secret must be a string and in
                BearerToken format

        Returns:
            - api_response: If the method is called asynchronously, returns the request thread

        Raises:
            - ValueError: if `pod_name` is `None` or `container_name` is `None`
            - TypeError: `exec_command` is not a list
        """
        if not pod_name or not container_name:
            raise ValueError(
                "The name of a Kubernetes pod and container must be provided."
            )
        if type(exec_command) != list:
            raise TypeError("The command to exec must be provided as a list")

        kube_kwargs = {**self.kube_kwargs, **(kube_kwargs or {})}

        api_client = cast(
            client.CoreV1Api, get_kubernetes_client("pod", kubernetes_api_key_secret)
        )

        api_response = stream(
            api_client.connect_get_namespaced_pod_exec,
            name=pod_name,
            namespace=namespace,
            container=container_name,
            command=exec_command,
            stderr=True,
            stdin=True,
            stdout=True,
            tty=False,
            **kube_kwargs
        )
        return api_response



from sodasql.scan.scan_builder import ScanBuilder

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from typing import Dict, Union


class SodaSQLScan(Task):
    """
    Task for running a SodaSQL scan given a scan definition and a warehouse definition.

    Args:
        - scan_def (dict, str, optional): scan definition.
            Can be either a path a SodaSQL Scan YAML file or a dictionary.
            For more information regarding SodaSQL Scan YAML files
            refer to https://docs.soda.io/soda-sql/documentation/scan.html
        - warehouse_def (dict, str, optional): warehouse definition.
            Can be either a path to a SodaSQL Warehouse YAML file or a dictionary.
            For more information regarding SodaSQL Warehouse YAML files
            refer to https://docs.soda.io/soda-sql/documentation/warehouse.html
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        scan_def: Union[Dict, str] = None,
        warehouse_def: Union[Dict, str] = None,
        **kwargs
    ):
        self.scan_def = scan_def
        self.warehouse_def = warehouse_def
        super().__init__(**kwargs)

    @classmethod
    def _get_scan_builder(
        cls, scan_def: Union[Dict, str], warehouse_def: Union[Dict, str]
    ) -> ScanBuilder:
        """
        Convenience method to build a SodaSQL ScanBuilder object

        Args:
            - scan_def (dict, str, optional): scan definition.
                Can be either a path a SodaSQL Scan YAML file or a dictionary.
                For more information regarding SodaSQL Scan YAML files
                refer to https://docs.soda.io/soda-sql/documentation/scan.html
            - warehouse_def (dict, str, optional): warehouse definition.
                Can be either a path to a SodaSQL Warehouse YAML file or a dictionary.
                For more information regarding SodaSQL Warehouse YAML files
                refer to https://docs.soda.io/soda-sql/documentation/warehouse.html

        Returns:
            - a SodaSQL ScanBuilder object
        """
        scan_builder = ScanBuilder()

        if isinstance(scan_def, str):
            scan_builder.scan_yml_file = scan_def
        elif isinstance(scan_def, dict):
            scan_builder.scan_yml_dict = scan_def

        if isinstance(warehouse_def, str):
            scan_builder.warehouse_yml_file = warehouse_def
        elif isinstance(warehouse_def, dict):
            scan_builder.warehouse_yml_dict = warehouse_def

        return scan_builder

    @defaults_from_attrs("scan_def", "warehouse_def")
    def run(
        self, scan_def: Union[Dict, str] = None, warehouse_def: Union[Dict, str] = None
    ):
        """
        Task run method. Execute a Scan against a Scan definition using a Warehouse definition.

        Args:
            - scan_def (dict, str, optional): scan definition.
                Can be either a path a SodaSQL Scan YAML file or a dictionary.
                For more information regarding SodaSQL Scan YAML files
                refer to https://docs.soda.io/soda-sql/documentation/scan.html
            - warehouse_def (dict, str, optional): warehouse definition.
                Can be either a path to a SodaSQL Warehouse YAML file or a dictionary.
                For more information regarding SodaSQL Warehouse YAML files
                refer to https://docs.soda.io/soda-sql/documentation/warehouse.html

        Returns:
            - A SodaSQL ScanResult object that contains all details regarding the execution
                of the scan (i.e: computed metrics, tests, errors, failures)

        Raises:
            - ValueError: if either scan_def or warehouse_def are None

        """
        if scan_def is None:
            raise ValueError(
                "Scan definition cannot be None. \
                Please provide either a path to a scan definition \
                file or a scan definition dictionary"
            )
        if warehouse_def is None:
            raise ValueError(
                "Warehouse definition cannot be None. \
                Please provide either a path to a warehouse definition \
                file or a warehouse definition dictionary"
            )
        scan_builder = self._get_scan_builder(
            scan_def=scan_def, warehouse_def=warehouse_def
        )
        scan = scan_builder.build()
        result = scan.execute()
        return result



from typing import Any

import prefect
from prefect import Task
from prefect.client import Client
from prefect.utilities.tasks import defaults_from_attrs


class RenameFlowRun(Task):
    """
    Task used to rename a running flow.

    Args:
        - flow_run_id (str, optional): The ID of the flow run to rename.
        - flow_run_name (str, optional): The new flow run name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        flow_run_id: str = None,
        flow_run_name: str = None,
        **kwargs: Any,
    ):
        self.flow_run_id = flow_run_id
        self.flow_run_name = flow_run_name
        super().__init__(**kwargs)

    @defaults_from_attrs("flow_run_id", "flow_run_name")
    def run(self, flow_run_id: str, flow_run_name: str) -> bool:
        """
        Args:
            - flow_run_id (str, optional): The ID of the flow run to rename. If `None`,
                the `flow_run_id` from `prefect.context` will be used as default value
            - flow_run_name (str, optional): The new flow run name

        Returns:
            - bool: Boolean representing whether the flow run was renamed successfully or not.

        Raises:
            - ValueError: If `flow_run_id` is not provided and `flow_run_id` does not exist
                in `prefect.context`
            - ValueError: If `flow_run_name` is not provided

        Example:
            ```python
            from prefect.tasks.prefect.flow_rename import FlowRenameTask

            rename_flow = FlowRenameTask(flow_name="A new flow run name")
            ```
        """
        flow_run_id = flow_run_id or prefect.context.get("flow_run_id")
        if not flow_run_id:
            raise ValueError(
                "`flow_run_id` must be explicitly provided or available in the context"
            )
        if flow_run_name is None:
            raise ValueError("Must provide a flow name.")

        client = Client()
        return client.set_flow_run_name(flow_run_id, flow_run_name)



import os

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from .hightouch_utils import HightouchClient


class HightouchRunSync(Task):
    """
    This task can be used to trigger a new sync run on [Hightouch](https://hightouch.io/).
    Under the hood
    it uses [Hightouch official APIs](https://hightouch.io/docs/syncs/api/#the-hightouch-rest-api).

    Args:
        - api_key (str, optional): The API key to use to authenticate on Hightouch.
        - api_key_env_var (str, optional): The name of the environment variable that contains
            the API key to use to authenticate on Hightouch.
            This value is considered only if `api_key` is not provided.
        - sync_id (int, optional): The ID of the Hightouch sync to run.
        - wait_for_completion (bool, optional): Whether to wait for sync run completion or not.
            Default is `False`.
        - wait_for_completion: (int, optional): Whether to wait for the sync run to finish execution
            or not.
            Default is `False`.
        - wait_time_between_api_calls (int, optional): The number of seconds to
            wait between API calls.
            Default is 10.
        - max_wait_time (int, optional): The maximum number of seconds to wait for the
            Hightouch API to return a response.
            If `wait_for_completion` is `True` and `maximum_wait_time` is not set then the
            task will run until completed or cancelled.
        - **kwargs (optional): Additional keyword arguments to pass to the
            standard Task initalization.
    """

    def __init__(
        self,
        api_key: str = None,
        api_key_env_var: str = None,
        sync_id: int = None,
        wait_for_completion: bool = False,
        wait_time_between_api_calls: int = 10,
        max_wait_time: int = None,
        **kwargs,
    ):
        self.api_key = api_key
        self.api_key_env_var = api_key_env_var
        self.sync_id = sync_id
        self.wait_for_completion = wait_for_completion
        self.wait_time_between_api_calls = wait_time_between_api_calls
        self.max_wait_time = max_wait_time
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "api_key",
        "api_key_env_var",
        "sync_id",
        "wait_for_completion",
        "wait_time_between_api_calls",
        "max_wait_time",
    )
    def run(
        self,
        api_key: str = None,
        api_key_env_var: str = None,
        sync_id: int = None,
        wait_for_completion: bool = False,
        wait_time_between_api_calls: int = 10,
        max_wait_time: int = None,
    ):
        """
        This task can be used to trigger a new sync run on [Hightouch](https://hightouch.io/).
        Under the hood it uses
        [Hightouch official APIs](https://hightouch.io/docs/syncs/api/#the-hightouch-rest-api).

        Args:
            - api_key (str, optional): The API key to use to authenticate on Hightouch.
            - api_key_env_var (str, optional): The name of the environment variable that contains
                the API key to use to authenticate on Hightouch.
                This value is considered only if `api_key` is not provided.
            - sync_id (int, optional): The ID of the Hightouch sync to run.
            - wait_for_completion: (int, optional): Whether to wait for the sync run to finish execution
                or not.
                Default is `False`.
            - wait_time_between_api_calls (int, optional): The number of seconds to
                wait between API calls.
                Default is 10.
            - max_wait_time (int, optional): The maximum number of seconds to wait for the
                If `wait_for_completion` is `True` and `maximum_wait_time` is not set then the
                task will run until completed or cancelled.
                Hightouch API to return a response.

        Raises:
            - `ValueError` if both `api_key` and `api_key_env_var` are missing.
            - `ValueError` if `api_key` is missing and `api_key_env_var` is not found.
            - `ValueError` if `sync_id` is missing.

        Returns:
            - if `wait_for_completion` is `False`, returns the response of
                [Start a new sync run](https://hightouch.io/docs/syncs/api/#start-a-new-sync-run).
            - if `wait_for_completion` is `True`, returns the response of
                [Get sync run status](https://hightouch.io/docs/syncs/api/#get-the-status-of-a-sync-run).
        """
        if not api_key and not api_key_env_var:
            msg = """Neither `api_key` nor `api_key_env_var` have been provided.
            Please provide either an API key or an environment variable
            where the API is stored."""
            raise ValueError(msg)

        if not api_key and api_key_env_var not in os.environ.keys():
            msg = f"""We were unable to find the environment variable ${api_key_env_var}.
            Please make sure the provided environment variable has been
            set with your Hightouch API key."""
            raise ValueError(msg)

        key = api_key if api_key else os.environ[api_key_env_var]

        if not sync_id:
            msg = """`sync_id` has not been provided. Please provide the ID for the
            Hightouch sync that you would like to trigger."""
            raise ValueError(msg)

        wait_api_call_secs = (
            wait_time_between_api_calls if wait_time_between_api_calls > 0 else 10
        )

        ht_client = HightouchClient(api_key=key)
        return ht_client.start_sync_run(
            sync_id=sync_id,
            wait_for_completion=wait_for_completion,
            wait_time_between_api_calls=wait_api_call_secs,
            max_wait_time=max_wait_time,
        )



"""
This module contains a collection of tasks to interact with Transform metrics layer.
"""

try:
    from prefect.tasks.transform.transform_tasks import TransformCreateMaterialization
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.transform` requires Prefect to be installed with the "transform" extra.'
    ) from err

__all__ = ["TransformCreateMaterialization"]



from typing import Any, Dict, List
from prefect.tasks.monte_carlo.client import MonteCarloClient

import prefect
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class MonteCarloIncorrectTagsFormat(Exception):
    """Exception for incorrect tags format"""

    pass


class MonteCarloCreateOrUpdateLineage(Task):
    """
    Task for creating or updating a lineage node for the given source and destination nodes,
    as well as for creating a lineage edge between those nodes.

    Args:
        - source (dict, optional): a source node configuration. Expected to include the following
            keys: `node_name`, `object_id`, `object_type`, `resource_name`, and optionally also
            a list of `tags`.
        - destination (dict, optional): a destination node configuration. Expected to include
            the following keys: `node_name`, `object_id`, `object_type`, `resource_name`,
            and optionally also a list of `tags`.
        - api_key_id (string, optional): to use this task,
            you need to pass the Monte Carlo API credentials.
            Those can be created using https://getmontecarlo.com/settings/api.
            To avoid passing the credentials
            in plain text, you can leverage the PrefectSecret task in your flow.
        - api_token (string, optional): the API token.
             To avoid passing the credentials in plain text,
             you can leverage PrefectSecret task in your flow.
        - prefect_context_tags (bool, optional): whether to automatically add
            tags with Prefect context.
        - expire_at (string, optional): date and time indicating when to expire
            a source-destination edge. You can expire specific lineage nodes.
            If this value is set, the
            edge between a source and destination nodes will expire on a given date.
            Expected format: "YYYY-MM-DDTHH:mm:ss.SSS". For example, "2042-01-01T00:00:00.000".
        - **kwargs (dict, optional): additional keyword arguments to pass
            to the Task constructor.

        Example source:

        ```python
        source = dict(
            # this may be a shorter version of object_id
            node_name="example_table_name",
            object_id="example_table_name",
            # "table" is recommended, but you can use any string, e.g. "csv_file"
            object_type="table",
            resource_name="name_your_source_system",
            tags=[{"propertyName": "prefect_test_key", "propertyValue": "prefect_test_value"}]
        )
        ```

        Example destination:

        ```python
        destination = dict(
            # the full name of a data warehouse table
            node_name="db_name:schema_name.table_name",
            object_id="db_name:schema_name.table_name",
            # "table" is recommended, but you can use any string, e.g. "csv_file"
            object_type="table",
            resource_name="your_dwh_resource_name",
            tags=[{"propertyName": "prefect_test_key", "propertyValue": "prefect_test_value"}]
        )
        ```

    """

    catalog_url = "https://getmontecarlo.com/catalog"

    def __init__(
        self,
        source: Dict[str, str] = None,
        destination: Dict[str, str] = None,
        api_key_id: str = None,
        api_token: str = None,
        prefect_context_tags: bool = True,
        expire_at: str = None,
        **kwargs: Any,
    ):
        self.source = source
        self.destination = destination
        self.api_key_id = api_key_id
        self.api_token = api_token
        self.prefect_context_tags = prefect_context_tags
        self.expire_at = expire_at
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "source",
        "destination",
        "api_key_id",
        "api_token",
        "prefect_context_tags",
        "expire_at",
    )
    def run(
        self,
        source: Dict[str, Any] = None,
        destination: Dict[str, Any] = None,
        api_key_id: str = None,
        api_token: str = None,
        prefect_context_tags: bool = True,
        expire_at: str = None,
    ) -> Dict[str, Any]:
        """
        Creates or updates a lineage node for the given source and destination,
        and creates a lineage edge between them.
        If a list of `tags` is set for either source or destination (or both),
        the task adds those to the source and/or destination nodes.
        If no `tags` are set, the task
        creates a node without custom tags. Additionally,
        if the `prefect_context_tags` flag is set to True (default),
        the task creates tags with information from Prefect context
        for debugging data downtime issues.

        Args:
            - source (dict, optional): a source node configuration.
                Expected to include the following keys: `node_name`,
                `object_id`, `object_type`, `resource_name`, and optionally also
                a list of `tags`.
            - destination (dict, optional): a destination node configuration.
                Expected to include the following keys:
                `node_name`, `object_id`, `object_type`, `resource_name`,
                and optionally also a list of `tags`.
            - api_key_id (string, optional): to use this task, you need to pass
                the Monte Carlo API credentials.
                Those can be created using https://getmontecarlo.com/settings/api.
                To avoid passing the credentials
                in plain text, you can leverage the PrefectSecret task in your flow.
            - api_token (string, optional): the API token.
                 To avoid passing the credentials in plain text,
                 you can leverage PrefectSecret task in your flow.
            - prefect_context_tags (bool, optional): whether to automatically add
                tags with Prefect context.
            - expire_at (string, optional): date and time indicating when to expire
                a source-destination edge. You can expire specific lineage nodes.
                If this value is set, the
                edge between a source and destination nodes will expire on a given date.
                Expected format: "YYYY-MM-DDTHH:mm:ss.SSS". For example, "2042-01-01T00:00:00.000".

        Returns:
            - dict: a GraphQL response dictionary with the edge ID.
        """
        if api_key_id is None:
            raise ValueError("Value for `api_key_id` must be provided.")

        if api_token is None:
            raise ValueError("Value for `api_token` must be provided.")

        if source.get("node_name") is None or destination.get("node_name") is None:
            raise ValueError(
                "Must provide a `node_name` in both source and destination."
            )

        if source.get("object_id") is None or destination.get("object_id") is None:
            raise ValueError(
                "Must provide an `object_id` in both source and destination."
            )

        if source.get("object_type") is None:
            source["object_type"] = "table"

        if destination.get("object_type") is None:
            destination["object_type"] = "table"

        if (
            source.get("resource_name") is None
            or destination.get("resource_name") is None
        ):
            raise ValueError(
                "Must provide a `resource_name` in both source and destination."
            )

        mc = MonteCarloClient(api_key_id, api_token)

        # source node
        if source.get("tags") is None and prefect_context_tags is False:
            source_node_mcon = mc.create_or_update_lineage_node(
                source["node_name"],
                source["object_id"],
                source["object_type"],
                source["resource_name"],
            )
        elif source.get("tags") is None and prefect_context_tags is True:
            source_node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                source["node_name"],
                source["object_id"],
                source["object_type"],
                source["resource_name"],
                tags=get_prefect_context_tags(),
            )
        elif source.get("tags") is not None and prefect_context_tags is False:
            source_node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                source["node_name"],
                source["object_id"],
                source["object_type"],
                source["resource_name"],
                source["tags"],
            )
        else:
            lineage_tags = [*source["tags"], *get_prefect_context_tags()]
            source_node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                source["node_name"],
                source["object_id"],
                source["object_type"],
                source["resource_name"],
                lineage_tags,
            )
        source_node_url = f"{self.catalog_url}/{source_node_mcon}/table"
        self.logger.info("Created or updated a source lineage node %s", source_node_url)

        # destination node
        if destination.get("tags") is None and prefect_context_tags is False:
            destination_node_mcon = mc.create_or_update_lineage_node(
                destination["node_name"],
                destination["object_id"],
                destination["object_type"],
                destination["resource_name"],
            )
        elif destination.get("tags") is None and prefect_context_tags is True:
            destination_node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                destination["node_name"],
                destination["object_id"],
                destination["object_type"],
                destination["resource_name"],
                tags=get_prefect_context_tags(),
            )
        elif destination.get("tags") is not None and prefect_context_tags is False:
            destination_node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                destination["node_name"],
                destination["object_id"],
                destination["object_type"],
                destination["resource_name"],
                destination["tags"],
            )
        else:
            lineage_tags = [*destination["tags"], *get_prefect_context_tags()]
            destination_node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                destination["node_name"],
                destination["object_id"],
                destination["object_type"],
                destination["resource_name"],
                lineage_tags,
            )
        destination_node_url = f"{self.catalog_url}/{destination_node_mcon}/table"
        self.logger.info(
            "Created or updated a destination lineage node %s", destination_node_url
        )

        # edge between source and destination nodes
        edge_id = mc.create_or_update_lineage_edge(source, destination, expire_at)
        self.logger.info("Created or updated a destination lineage edge %s", edge_id)
        return edge_id


class MonteCarloCreateOrUpdateNodeWithTags(Task):
    """
    Task for creating or updating a lineage node in the Monte Carlo Catalog
    and enriching it with multiple metadata tags.
    If the `prefect_context_tags` flag is set to True (default),
    the task additionally creates tags with Prefect
    context for debugging data downtime issues.

    Args:
        - node_name (string, optional): the display name of a lineage node.
        - object_id (string, optional): the object ID of a lineage node.
        - object_type (string, optional): the object type of a lineage node - usually,
            either "table" or "view".
        - resource_name (string, optional): name of the data warehouse or custom resource.
            All resources can be retrieved via a separate task.
        - lineage_tags (list, optional): a list of dictionaries where each dictionary is a single tag.
        - api_key_id (string, optional): to use this task, you need to pass
            the Monte Carlo API credentials.
            Those can be created using https://getmontecarlo.com/settings/api.
            To avoid passing the credentials
            in plain text, you can leverage the PrefectSecret task in your flow.
        - api_token (string, optional): the API token.
             To avoid passing the credentials in plain text,
             you can leverage the PrefectSecret task in your flow.
        - prefect_context_tags (bool, optional): whether to automatically add
            tags with Prefect context.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor.

        The expected format for `lineage_tags`:

        ```python
        [{"propertyName": "tag_name", "propertyValue": "your_tag_value"},
        {"propertyName": "another_tag_name", "propertyValue": "your_tag_value"}]
        ```
    """

    def __init__(
        self,
        node_name: str = None,
        object_id: str = None,
        object_type: str = None,
        resource_name: str = None,
        lineage_tags: List[Dict[str, str]] = None,
        api_key_id: str = None,
        api_token: str = None,
        prefect_context_tags: bool = True,
        **kwargs: Any,
    ):
        self.node_name = node_name
        self.object_id = object_id
        self.object_type = object_type
        self.resource_name = resource_name
        self.lineage_tags = lineage_tags
        self.api_key_id = api_key_id
        self.api_token = api_token
        self.prefect_context_tags = prefect_context_tags
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "node_name",
        "object_id",
        "object_type",
        "resource_name",
        "lineage_tags",
        "api_key_id",
        "api_token",
        "prefect_context_tags",
    )
    def run(
        self,
        node_name: str = None,
        object_id: str = None,
        object_type: str = None,
        resource_name: str = None,
        lineage_tags: List[Dict[str, str]] = None,
        api_key_id: str = None,
        api_token: str = None,
        prefect_context_tags: bool = True,
    ) -> str:
        """
        Creates or updates multiple metadata tags for a given lineage node in the Monte Carlo Catalog.
        If the `prefect_context_tags` flag is set to True (default), the task additionally creates
        a tag with the Prefect context dictionary for debugging data downtime issues.

        Args:
            - node_name (string, optional): the display name of a lineage node.
            - object_id (string, optional): the object ID of a lineage node.
            - object_type (string, optional): the object type of a lineage node - usually,
                either "table" or "view".
            - resource_name (string, optional): name of the data warehouse or custom resource.
                All resources can be retrieved via a separate task.
            - lineage_tags (list, optional): a list of dictionaries
                where each dictionary is a single tag.
            - api_key_id (string, optional): to use this task,
                you need to pass the Monte Carlo API credentials.
                Those can be created using https://getmontecarlo.com/settings/api.
                To avoid passing the credentials
                in plain text, you can leverage the PrefectSecret task in your flow.
            - api_token (string, optional): the API token.
                 To avoid passing the credentials in plain text,
                 you can leverage the PrefectSecret task in your flow.
            - prefect_context_tags (bool, optional): whether to automatically add a tag
                with Prefect context.

        Returns:
            - string: MCON - a Monte Carlo internal ID of the node.
        """
        if node_name is None:
            raise ValueError("Value for `node_name` must be provided.")
        if object_id is None:
            raise ValueError("Value for `object_id` must be provided.")
        if object_type is None:
            raise ValueError("Value for `object_type` must be provided.")
        if resource_name is None:
            raise ValueError("Value for `resource_name` must be provided.")
        if api_key_id is None:
            raise ValueError("Value for `api_key_id` must be provided.")
        if api_token is None:
            raise ValueError("Value for `api_token` must be provided.")

        mc = MonteCarloClient(api_key_id, api_token)
        if lineage_tags is None and prefect_context_tags is False:
            node_mcon = mc.create_or_update_lineage_node(
                node_name,
                object_id,
                object_type,
                resource_name,
            )
        else:
            for tag in lineage_tags:
                if sorted(tag.keys()) != ["propertyName", "propertyValue"]:
                    raise MonteCarloIncorrectTagsFormat(
                        "Must provide tags in the format "
                        '[{"propertyName": "tag_name", "propertyValue": "tag_value"}].',
                        "You provided: ",
                        tag,
                    )
            if prefect_context_tags is False:
                node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                    node_name,
                    object_id,
                    object_type,
                    resource_name,
                    lineage_tags,
                )
            else:
                final_tags = [*lineage_tags, *get_prefect_context_tags()]
                node_mcon = mc.create_or_update_lineage_node_with_multiple_tags(
                    node_name, object_id, object_type, resource_name, tags=final_tags
                )
        return node_mcon


class MonteCarloGetResources(Task):
    """
    Task for querying resources using the Monte Carlo API.
    You can use this task to find the
    name of your data warehouse or other resource, and use it as `resource_name`
    in other Monte Carlo tasks.

    Args:
        - api_key_id (string, optional): to use this task, you need to pass
            the Monte Carlo API credentials.
            Those can be created using https://getmontecarlo.com/settings/api.
            To avoid passing the credentials
            in plain text, you can leverage the PrefectSecret task in your flow.
        - api_token (string, optional): the API token.
             To avoid passing the credentials in plain text, you can leverage the PrefectSecret task
             in your flow.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        api_key_id: str = None,
        api_token: str = None,
        **kwargs: Any,
    ):
        self.api_key_id = api_key_id
        self.api_token = api_token
        super().__init__(**kwargs)

    @defaults_from_attrs("api_key_id", "api_token")
    def run(
        self,
        api_key_id: str = None,
        api_token: str = None,
    ) -> List[Dict[str, Any]]:
        """
        Retrieve all Monte Carlo resources.

        Args:
            - api_key_id (string, optional): to use this task,
                you need to pass the Monte Carlo API credentials.
                Those can be created using https://getmontecarlo.com/settings/api.
                To avoid passing the credentials
                in plain text, you can leverage the PrefectSecret task in your flow.
            - api_token (string, optional): the API token.
                 To avoid passing the credentials in plain text, you can leverage the PrefectSecret
                 task in your flow.

        Returns:
            - list: Monte Carlo resources, including all existing data warehouses and custom resources.
        """
        if api_key_id is None:
            raise ValueError("Value for `api_key_id` must be provided.")
        if api_token is None:
            raise ValueError("Value for `api_token` must be provided.")
        mc = MonteCarloClient(api_key_id, api_token)
        return mc.get_resources()


def get_prefect_context_tags() -> List[Dict[str, Any]]:
    """
    Helper function to generate a list of Prefect context tags
    when `prefect_context_tags` is set to True.

    Returns:
        - list: a list of dictionaries with the task's context.
    """
    return [
        dict(propertyName="flow_id", propertyValue=str(prefect.context.get("flow_id"))),
        dict(
            propertyName="flow_name",
            propertyValue=str(prefect.context.get("flow_name")),
        ),
        dict(
            propertyName="flow_run_id",
            propertyValue=str(prefect.context.get("flow_run_id")),
        ),
        dict(
            propertyName="flow_run_name",
            propertyValue=str(prefect.context.get("flow_run_name")),
        ),
        dict(
            propertyName="flow_run_version",
            propertyValue=str(prefect.context.get("flow_run_version")),
        ),
        dict(
            propertyName="project_id",
            propertyValue=str(prefect.context.get("project_id")),
        ),
        dict(
            propertyName="project_name",
            propertyValue=str(prefect.context.get("project_name")),
        ),
        dict(
            propertyName="scheduled_start_time",
            propertyValue=str(prefect.context.get("scheduled_start_time")),
        ),
        dict(
            propertyName="execution_start_time",
            propertyValue=str(prefect.context.get("date")),
        ),
        dict(propertyName="task_id", propertyValue=str(prefect.context.get("task_id"))),
        dict(
            propertyName="task_run_id",
            propertyValue=str(prefect.context.get("task_run_id")),
        ),
        dict(
            propertyName="task_run_name",
            propertyValue=str(prefect.context.get("task_run_name")),
        ),
        dict(
            propertyName="task_slug",
            propertyValue=str(prefect.context.get("task_slug")),
        ),
        dict(
            propertyName="task_name",
            propertyValue=str(prefect.context.get("task_name")),
        ),
        dict(
            propertyName="task_full_name",
            propertyValue=str(prefect.context.get("task_full_name")),
        ),
        dict(
            propertyName="task_run_count",
            propertyValue=str(prefect.context.get("task_run_count")),
        ),
        dict(
            propertyName="map_index",
            propertyValue=str(prefect.context.get("map_index")),
        ),
        dict(
            propertyName="parameters",
            propertyValue=str(prefect.context.get("parameters")),
        ),
        dict(
            propertyName="running_with_backend",
            propertyValue=str(prefect.context.get("running_with_backend")),
        ),
    ]



import os
import re
from typing import TYPE_CHECKING, Any
from urllib.error import HTTPError

import requests
from requests.exceptions import HTTPError as RequestsHTTPError
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

import prefect
from prefect.client import Secret
from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file


BITBUCKET_CLOUD_RETRY = Retry(
    total=10,
    backoff_factor=0.5,
    status_forcelist=[429],
    # typeshed is out of date with urllib3 and missing `allowed_methods`
    allowed_methods=["HEAD", "GET", "OPTIONS"],  # type: ignore
)

if TYPE_CHECKING:
    import atlassian
    from prefect.core.flow import Flow


class Bitbucket(Storage):
    """
    Bitbucket storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a Bitbucket repository. This is for Bitbucket Server or Bitbucket Cloud.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has Bitbucket storage:

    ```python
    flow = Flow("my-flow")
    flow.storage = Bitbucket(
        project="my.project", repo="my.repo", path="/flows/flow.py", ref="my-branch"
    )
    ```

    - Push this `flow.py` file to the `my.repo` repository under `/flows/flow.py` inside "my.project"
        project.

    - Call `prefect register -f flow.py` to register this flow with Bitbucket storage.

    Args:
        - project (str): project that the repository will be in.  Not equivalent to a GitHub project;
            required value for all Bitbucket repositories.
        - repo (str): the repository name, with complete taxonomy.
        - workspace (str, optional): the workspace name. Bitbucket cloud only.
        - host (str, optional): the server host. Bitbucket server only.
        - path (str, optional): a path pointing to a flow file in the repo
        - ref (str, optional): a commit SHA-1 value or branch name or tag. If not specified,
            defaults to master branch for the repo.
        - access_token_secret (str, optional): the name of a Prefect secret
            that contains a Bitbucket access token to use when loading flows from
            this storage. Bitbucket Server only
        - cloud_username_secret (str, optional): the name of a Prefect secret that contains a
            Bitbucket username to use when loading flows from this storage. Bitbucket Cloud only.
        - cloud_app_password_secret (str, optional): the name of a Prefect secret that contains a
            Bitbucket app password, from the account associated with the `cloud_username_secret`,
            to use when loading flows from this storage. Bitbucket Cloud only.
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        project: str,
        repo: str,
        workspace: str = None,
        host: str = None,
        path: str = None,
        ref: str = None,
        access_token_secret: str = None,
        cloud_username_secret: str = None,
        cloud_app_password_secret: str = None,
        **kwargs: Any,
    ) -> None:
        self.project = project
        self.repo = repo
        self.workspace = workspace
        self.host = host
        self.path = path
        self.ref = ref
        self.access_token_secret = access_token_secret
        self.cloud_username_secret = cloud_username_secret
        self.cloud_app_password_secret = cloud_app_password_secret

        if self.workspace and self.access_token_secret:
            raise ValueError(
                "`access_token_secret` and `workspace` are both set. Setting `workspace` "
                "is only for use with Bitbucket Cloud which uses `cloud_username_secret` "
                "and `cloud_app_password_secret` to provide authentication."
            )

        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """

        return (
            self._get_flow_from_bitbucket_cloud(flow_name)
            if self.workspace
            else self._get_flow_from_bitbucket_server(flow_name)
        )

    def _get_flow_from_bitbucket_cloud(self, flow_name: str) -> "Flow":
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        # Use ref attribute if present, defaulting to "master"
        ref = self.ref or "master"

        client = self._get_bitbucket_cloud_client()

        try:

            # Fetch SHA-1 of all branches if ref is not alread a SHA-1 hash
            SHA1_REGEX = r"^[0-9a-f]{40}$"
            if not re.match(SHA1_REGEX, ref):
                branches = client.get(
                    f"repositories/{self.workspace}/{self.repo}/refs/branches"
                )
                branch_hashes = {
                    branch["name"]: branch["target"]["hash"]
                    for branch in branches["values"]
                }

                if ref in branch_hashes:
                    ref = branch_hashes[ref]

            # Fetch SHA-1 of all tags if ref is not alread a SHA-1 hash
            if not re.match(SHA1_REGEX, ref):
                tags = client.get(
                    f"repositories/{self.workspace}/{self.repo}/refs/tags"
                )
                tag_hashes = {
                    tag["name"]: tag["target"]["hash"] for tag in tags["values"]
                }

                if ref in tag_hashes:
                    ref = tag_hashes[ref]

            if not re.match(SHA1_REGEX, ref):
                raise ValueError(
                    f"ref {ref!r} not found in '{self.workspace}/{self.repo}'"
                )

            contents_url = (
                f"repositories/{self.workspace}/{self.repo}/src/{ref}/{flow_location}"
            )
            self.logger.info(f"Downloading flow from Bitbucket cloud - {contents_url}")
            contents = client.get(contents_url)

        except RequestsHTTPError as err:

            if err.response.status_code == 403:
                err_msg = "Access denied to repository. Please check credentials."
            else:
                err_msg = (
                    "Error retrieving contents at "
                    f"{flow_location} in {self.workspace}/{self.project}/{self.repo}@{ref}. "
                    "Please check arguments passed to Bitbucket storage."
                )

            self.logger.error(err_msg)
            raise RuntimeError(err_msg) from err

        return extract_flow_from_file(file_contents=contents, flow_name=flow_name)

    def _get_flow_from_bitbucket_server(self, flow_name: str) -> "Flow":
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        # Use ref attribute if present, defaulting to "master"
        ref = self.ref or "master"

        client = self._get_bitbucket_server_client()

        try:
            contents = client.get_content_of_file(
                self.project,
                self.repo,
                flow_location,
                at=ref,
            )
        except HTTPError as err:
            if err.code == 401:
                self.logger.error(
                    "Access denied to repository. Please check credentials."
                )
                raise
            elif err.code == 404:
                self.logger.error(
                    "Invalid address. Check that host, project, and repository are correct."
                )
                raise
            else:
                self.logger.error(
                    f"Error retrieving contents at {flow_location} in {self.repo}@{ref}. "
                    "Please check arguments passed to Bitbucket storage and verify project exists."
                )
                raise

        return extract_flow_from_file(file_contents=contents, flow_name=flow_name)

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    def _get_bitbucket_server_client(self) -> "atlassian.Bitbucket":
        try:
            from atlassian import Bitbucket
        except ImportError as exc:
            raise ImportError(
                "Unable to import atlassian, please ensure you have installed the bitbucket extra"
            ) from exc

        if self.access_token_secret is not None:
            # If access token secret specified, load it
            access_token = Secret(self.access_token_secret).get()
        else:
            # Otherwise, fallback to loading from local secret or environment variable
            access_token = prefect.context.get("secrets", {}).get(
                "BITBUCKET_ACCESS_TOKEN"
            )
            if access_token is None:
                access_token = os.getenv("BITBUCKET_ACCESS_TOKEN")

        host = "https://bitbucket.org" if self.host is None else self.host

        session = requests.Session()
        if access_token is None:
            session.headers["Authorization"] = "Bearer "
        else:
            session.headers["Authorization"] = "Bearer " + access_token
        return Bitbucket(host, session=session)

    def _get_bitbucket_cloud_client(self) -> "atlassian.Bitbucket":
        try:
            from atlassian.bitbucket.cloud import Cloud
        except ImportError as exc:
            raise ImportError(
                "Unable to import atlassian, please ensure you have installed the bitbucket extra"
            ) from exc

        cloud_username = (
            Secret(self.cloud_username_secret).get()
            if self.cloud_username_secret is not None
            else os.getenv("BITBUCKET_CLOUD_USERNAME")
        )
        cloud_app_password = (
            Secret(self.cloud_app_password_secret).get()
            if self.cloud_app_password_secret is not None
            else os.getenv("BITBUCKET_CLOUD_APP_PASSWORD")
        )

        # Bitbucket clould api is rate limited, retry at this frequency
        # 0.25, 0.5, 1, 2, 4, 8, 16, 32, 64, 128
        adapter = HTTPAdapter(max_retries=BITBUCKET_CLOUD_RETRY)
        session = requests.Session()
        session.mount("https://", adapter)

        return Cloud(
            username=cloud_username,
            password=cloud_app_password,
            cloud=True,
            session=session,
        )



from typing import Dict, Optional

from prefect import Task
from prefect.tasks.databricks.databricks_hook import DatabricksHook
from prefect.utilities.tasks import defaults_from_attrs


class DatabricksGetJobID(Task):
    """
    Finds a job_id corresponding to a job name on Databricks using the
    [api/2.1/jobs/list](https://docs.databricks.com/dev-tools/api/latest/jobs.html#operation/JobsList)
    API endpoint.

    Args:
        - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks
            Connection String. Structure must be a string of valid JSON. To use token based
            authentication, provide the key `token` in the string for the connection and create the
            key `host`.
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
            OR
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
            See documentation of the `DatabricksSubmitRun` Task to see how to pass in the connection
            string using `PrefectSecret`.
        - search_limit (int, optional): Controls the number of jobs to return per API call,
            This value must be greater than 0 and less than or equal to 25.
        - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
            this run. By default the task will poll every 30 seconds.
        - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
            unreachable. Its value must be greater than or equal to 1.
        - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
            might be a floating point number).
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.

    Returns:
        - job_id (int): Job id of the job name.

    Examples:
        You can use the task to feed in the `job_id` for `DatabricksRunNow`.
        ```
        conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
        get_job_id = DatabricksGetJobID(databricks_conn_secret=conn)
        dbx_job_id = get_job_id(job_name="dbx")

        notebook_run = DatabricksRunNow(
            job_id=dbx_job_id,
            notebook_params=notebook_params,
            python_params=python_params,
            spark_submit_params=spark_submit_params,
            jar_params=jar_params
        )
        notebook_run(databricks_conn_secret=conn)
        ```
    """

    def __init__(
        self,
        databricks_conn_secret: Optional[Dict[str, str]] = None,
        search_limit: int = 25,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
        **kwargs
    ) -> None:
        self.databricks_conn_secret = databricks_conn_secret
        self.search_limit = search_limit
        self.polling_period_seconds = polling_period_seconds
        self.databricks_retry_limit = databricks_retry_limit
        self.databricks_retry_delay = databricks_retry_delay

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "databricks_conn_secret",
        "search_limit",
        "polling_period_seconds",
        "databricks_retry_limit",
        "databricks_retry_delay",
    )
    def run(
        self,
        job_name: str,
        databricks_conn_secret: Optional[Dict[str, str]] = None,
        search_limit: int = 25,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
    ) -> int:
        """
        Task run method.

        Args:
            - job_name (str): The job_name of an existing Databricks job.
            - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks
                Connection String. Structure must be a string of valid JSON. To use token based
                authentication, provide the key `token` in the string for the connection and create the
                key `host`.
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
                OR
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
                See documentation of the `DatabricksSubmitRun` Task to see how to pass in the connection
                string using `PrefectSecret`.
            - search_limit (int, optional): Controls the number of jobs to return per API call,
                This value must be greater than 0 and less or equal to 25.
                this run. By default the task will poll every 30 seconds.
            - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
                this run. By default the task will poll every 30 seconds.
            - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
                unreachable. Its value must be greater than or equal to 1.
            - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
                might be a floating point number).

        Returns:
            - job_id (int): Job id of the job name.
        """
        # Initialize Databricks Connections
        hook = self._get_hook()

        # Fetch Job ID
        self.logger.info("Searching for job_ids with name: %s ", job_name)
        job_id = hook.get_job_id_by_name(job_name=job_name, limit=search_limit)

        return job_id

    def _get_hook(self):
        # this is a function so it's easier to mock in testing
        return DatabricksHook(
            self.databricks_conn_secret,
            retry_limit=self.databricks_retry_limit,
            retry_delay=self.databricks_retry_delay,
        )



import os
import tempfile
import logging
import sys
from subprocess import PIPE, STDOUT, Popen
from typing import Any, List, Union, Optional

import prefect
from prefect.utilities.tasks import defaults_from_attrs


class PowershellTask(prefect.Task):
    """
    Task for running arbitrary Powershell commands.

    NOTE: This task combines stderr and stdout because reading from both
          streams without blocking is tricky.

    Args:
        - command (string, optional): shell command to be executed; can also be
            provided post-initialization by calling this task instance
        - env (dict, optional): dictionary of environment variables to use for
            the subprocess; can also be provided at runtime
        - helper_script (str, optional): a string representing a shell script, which
            will be executed prior to the `command` in the same process. Can be used to
            change directories, define helper functions, etc. when re-using this Task
            for different commands in a Flow; can also be provided at runtime
        - shell (string, optional): shell to run the command with; Can be
            `powershell` or `pwsh`; defaults to "pwsh". `pwsh` is the
            executable filename of PowerShell [Core] (v6+), while `powershell` is
            the executable name of Windows PowerShell (v5.1-)
        - return_all (bool, optional): boolean specifying whether this task
            should return all lines of stdout as a list, or just the last line
            as a string; defaults to `False`
        - log_stderr (bool, optional): boolean specifying whether this task should log
            the output in the case of a non-zero exit code; defaults to `False`. This
             actually logs both stderr and stdout and will only log the last line of
             output unless `return_all` is `True`
        - stream_output (Union[bool, int, str], optional): specifies whether this task should log
            the output as it occurs, and at what logging level. If `True` is passed,
            the logging level defaults to `INFO`; otherwise, any integer or string
            value that's passed will be treated as the log level, provided
            the `logging` library can successfully interpret it. If enabled,
            `log_stderr` will be ignored as the output will have already been
            logged. defaults to `False`
        - **kwargs: additional keyword arguments to pass to the Task constructor

    Raises:
        - TypeError: if `stream_output` is passed in as a string, but cannot
          successfully be converted to a numeric value by logging.getLevelName()

    Example:
        ```python
        from prefect import Flow
        from prefect.tasks.shell import PowershellTask

        task = PowershellTask(helper_script="cd ~")
        with Flow("My Flow") as f:
            # task will be executed in home directory
            contents = task(command='ls')

        out = f.run()
        ```
    """

    def __init__(
        self,
        command: str = None,
        env: dict = None,
        helper_script: str = None,
        shell: str = "pwsh",
        return_all: bool = False,
        log_stderr: bool = False,
        stream_output: Union[bool, int, str] = False,
        **kwargs: Any,
    ):
        self.command = command
        self.env = env
        self.helper_script = helper_script
        self.shell = shell
        self.return_all = return_all
        self.log_stderr = log_stderr

        if isinstance(stream_output, str):
            stream_output = logging.getLevelName(stream_output)
            if not isinstance(stream_output, int):
                raise TypeError(
                    f"'stream_output': {stream_output} is not a valid log level"
                )

        self.stream_output = logging.INFO if stream_output is True else stream_output
        super().__init__(**kwargs)

    @defaults_from_attrs("command", "env", "helper_script")
    def run(
        self, command: str = None, env: dict = None, helper_script: str = None
    ) -> Optional[Union[str, List[str]]]:
        """
        Run the shell command.

        Args:
            - command (string): shell command to be executed; can also be
                provided at task initialization. Any variables / functions defined in
                `helper_script` will be available in the same process this command
                runs in
            - env (dict, optional): dictionary of environment variables to use for
                the subprocess
            - helper_script (str, optional): a string representing a shell script, which
                will be executed prior to the `command` in the same process. Can be used to
                change directories, define helper functions, etc. when re-using this Task
                for different commands in a Flow

        Returns:
            - output (string): if `return_all` is `False` (the default), only
                the last line of output is returned, otherwise all lines are
                returned, which is useful for passing result of shell command
                to other downstream tasks. If there is no output, `None` is
                returned.

        Raises:
            - prefect.engine.signals.FAIL: if command has an exit code other
                than 0
        """
        if command is None:
            raise TypeError("run() missing required argument: 'command'")

        current_env = os.environ.copy()
        current_env.update(env or {})

        tmp = tempfile.NamedTemporaryFile(
            prefix="prefect-", suffix=".ps1", delete=False
        )
        if helper_script:
            tmp.write(helper_script.encode())
            tmp.write(os.linesep.encode())
        tmp.write(command.encode())
        tmp.close()

        with Popen(
            [self.shell, tmp.name],
            stdout=PIPE,
            stderr=STDOUT,
            env=current_env,
            # Windows does not use the PATH during subprocess creation
            # by default so we will use `shell` mode to do so
            shell=sys.platform == "win32",
        ) as sub_process:
            line = None
            lines = []
            # mypy: None is not compatible with non-optional types
            # https://mypy.readthedocs.io/en/latest/common_issues.html#unexpected-errors-about-none-and-or-optional-types
            assert sub_process.stdout is not None
            for raw_line in iter(sub_process.stdout.readline, b""):
                line = raw_line.decode("utf-8").rstrip()

                if self.return_all:
                    lines.append(line)

                if self.stream_output:
                    self.logger.log(level=self.stream_output, msg=line)

            sub_process.wait()
            if sub_process.returncode:
                msg = "Command failed with exit code {}".format(
                    sub_process.returncode,
                )

                self.logger.error(msg)

                if self.log_stderr and not self.stream_output:
                    self.logger.error("\n".join(lines) if self.return_all else line)

                raise prefect.engine.signals.FAIL(
                    msg,
                    result=lines if self.return_all else line,
                )

        os.unlink(tmp.name)
        return lines if self.return_all else line



import warnings

try:
    from prefect.tasks.database.sqlite import SQLiteQuery, SQLiteScript
except ImportError:
    warnings.warn(
        "SQLite tasks require sqlite3 to be installed", UserWarning, stacklevel=2
    )

__all__ = ["SQLiteQuery", "SQLiteScript"]
